'''
fsck.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import atexit
import itertools
import logging
import os
import re
import shutil
import stat
import sys
import textwrap
import time
from datetime import datetime
from os.path import basename
from tempfile import NamedTemporaryFile

import apsw

from . import CTRL_INODE, CURRENT_FS_REV, ROOT_INODE
from .backends.common import NoSuchObject
from .backends.comprenc import ComprencBackend
from .backends.local import Backend as LocalBackend
from .common import (
    get_backend,
    get_path,
    inode_for_path,
    is_mounted,
    sha256_fh,
    thaw_basic_mapping,
    time_ns,
)
from .database import (
    Connection,
    FsAttributes,
    NoSuchRowError,
    download_metadata,
    expire_objects,
    get_available_seq_nos,
    read_cached_params,
    read_remote_params,
    upload_metadata,
    upload_params,
    write_params,
)
from .logging import QuietError, setup_logging, setup_warnings
from .parse_args import ArgumentParser

log = logging.getLogger(__name__)

S_IFMT = (
    stat.S_IFDIR
    | stat.S_IFREG
    | stat.S_IFSOCK
    | stat.S_IFBLK
    | stat.S_IFCHR
    | stat.S_IFIFO
    | stat.S_IFLNK
)


class Fsck:
    def __init__(self, cachedir_, backend_, param, conn):
        self.cachedir = cachedir_
        self.backend = backend_
        self.expect_errors = False
        self.found_errors = False
        self.uncorrectable_errors = False
        self.max_obj_size = param.data_block_size
        self.conn = conn

        # Set of objects that have been unlinked by check_cache.
        # check_object_refcounts() will not report errors if these objects still
        # exist even though they have refcount=0
        self.unlinked_objects = set()

    def check(self, check_cache=True):
        """Check file system

        If *check_cache* is False, assume that all cache files are clean (aka
        have been uploaded to the backend). If *check_cache* is ``keep``, do not
        remove cache files on exit.

        Sets instance variable `found_errors`.
        """

        # Create indices required for reference checking
        log.info('Creating temporary extra indices...')
        for idx in ('tmp1', 'tmp2', 'tmp3', 'tmp4'):
            self.conn.execute('DROP INDEX IF EXISTS %s' % idx)
        self.conn.execute('CREATE INDEX tmp1 ON inode_blocks(obj_id)')
        self.conn.execute('CREATE INDEX tmp2 ON contents(inode)')
        self.conn.execute('CREATE INDEX tmp3 ON contents(name_id)')
        self.conn.execute('CREATE INDEX tmp4 ON ext_attributes(name_id)')
        try:
            self.check_lof()
            if check_cache:
                self.check_cache(check_cache == 'keep')
            self.check_names_refcount()

            self.check_contents_name()
            self.check_contents_inode()
            self.check_contents_parent_inode()

            self.check_objects_temp()
            self.check_objects_hash()
            self.check_objects_refcount()
            self.check_objects_id()
            self.check_objects_phys_size()
            # TODO: Are there any checks we could do on objects.length?

            self.check_inode_blocks_obj_id()
            self.check_inode_blocks_inode()

            self.check_inodes_refcount()
            self.check_inodes_size()

            self.check_ext_attributes_name()
            self.check_ext_attributes_inode()

            self.check_symlinks_inode()

            self.check_loops()
            self.check_unix()
            self.check_foreign_keys()
        finally:
            log.info('Dropping temporary indices...')
            for idx in ('tmp1', 'tmp2', 'tmp3', 'tmp4'):
                self.conn.execute('DROP INDEX %s' % idx)

    def log_error(self, *a, **kw):
        '''Log file system error if not expected'''

        if self.expect_errors:
            return log.info(*a, **kw)
        else:
            return log.warning(*a, **kw)

    def check_foreign_keys(self):
        '''Check for referential integrity

        Checks that all foreign keys in the SQLite tables actually resolve.
        This is necessary, because we disable runtime checking by SQLite
        for performance reasons.

        Note: any problems should have already been caught by the more
        specific checkers.
        '''

        log.info("Checking referential integrity...")

        for (table,) in self.conn.query("SELECT name FROM sqlite_master WHERE type='table'"):
            for row in self.conn.query('PRAGMA foreign_key_list(%s)' % table):
                sql_objs = {
                    'src_table': table,
                    'dst_table': row[2],
                    'src_col': row[3],
                    'dst_col': row[4],
                    'dbf': self.conn.fixes['dbf'],
                }

                for (val,) in self.conn.query(
                    'SELECT %(src_table)s.%(src_col)s '
                    'FROM %(src_table)s LEFT JOIN %(dst_table)s '
                    'ON %(dbf)s%(src_table)s.%(src_col)s = %(dbf)s%(dst_table)s.%(dst_col)s '
                    'WHERE %(dbf)s%(dst_table)s.%(dst_col)s IS NULL '
                    'AND %(src_table)s.%(src_col)s IS NOT NULL' % sql_objs
                ):
                    self.found_errors = True
                    sql_objs['val'] = val
                    self.log_error(
                        '%(src_table)s.%(src_col)s refers to non-existing key %(val)s '
                        'in %(dst_table)s.%(dst_col)s, deleting.',
                        sql_objs,
                    )
                    log.error('This should not happen, please report a bug.')
                    self.uncorrectable_errors = True

    def check_cache(self, keep_cache=False):
        """Commit uncommitted cache files"""

        log.info("Checking for dirty cache objects...")
        if not os.path.exists(self.cachedir):
            return
        candidates = os.listdir(self.cachedir)

        total = len(candidates)
        for i, filename in enumerate(candidates):
            i += 1  # start at 1

            log.info(
                'Processed %d/%d dirty cache objects (%d%%)',
                i,
                total,
                i / total * 100,
                extra={'rate_limit': 1, 'update_console': True, 'is_last': i == total},
            )

            match = re.match('^(\\d+)-(\\d+)$', filename)
            if match:
                inode = int(match.group(1))
                blockno = int(match.group(2))
            elif re.match('^(\\d+)-(\\d+)\\.tmp$', filename):
                # Temporary file created when downloading object
                self.found_errors = True
                self.log_error("Removing leftover temporary file: " + filename)
                os.unlink(os.path.join(self.cachedir, filename))
                continue
            else:
                raise RuntimeError('Strange file in cache directory: %s' % filename)

            # Calculate block checksum
            with open(os.path.join(self.cachedir, filename), "rb") as fh:
                size = os.fstat(fh.fileno()).st_size
                hash_should = sha256_fh(fh).digest()
            log.debug('%s has checksum %s', filename, hash_should)

            # Check if stored object has same checksum
            try:
                (obj_id, hash_is) = self.conn.get_row(
                    'SELECT obj_id, hash '
                    'FROM inode_blocks JOIN objects ON obj_id = objects.id '
                    'WHERE inode=? AND blockno=?',
                    (
                        inode,
                        blockno,
                    ),
                )
            except NoSuchRowError:
                pass
            else:
                if hash_should == hash_is:
                    if not keep_cache:
                        os.unlink(os.path.join(self.cachedir, filename))
                    continue

            self.found_errors = True
            self.log_error("Writing dirty block %d of inode %d to backend", blockno, inode)

            try:
                obj_id = self.conn.get_val('SELECT id FROM objects WHERE hash=?', (hash_should,))

            except NoSuchRowError:
                obj_id = self.conn.rowid(
                    'INSERT INTO objects (refcount, phys_size, length, hash) '
                    'VALUES(1, -1, ?, ?)',
                    (size, hash_should),
                )

                with open(os.path.join(self.cachedir, filename), "rb") as fh:
                    obj_size = self.backend.write_fh('s3ql_data_%d' % obj_id, fh)

                self.conn.execute('UPDATE objects SET phys_size=? WHERE id=?', (obj_size, obj_id))
            else:
                self.conn.execute('UPDATE objects SET refcount=refcount+1 WHERE id=?', (obj_id,))

            try:
                old_obj_id = self.conn.get_val(
                    'SELECT obj_id FROM inode_blocks WHERE inode=? AND blockno=?',
                    (inode, blockno),
                )
            except NoSuchRowError:
                self.conn.execute(
                    'INSERT INTO inode_blocks (obj_id, inode, blockno) VALUES(?,?,?)',
                    (obj_id, inode, blockno),
                )
            else:
                self.conn.execute(
                    'UPDATE inode_blocks SET obj_id=? WHERE inode=? AND blockno=?',
                    (obj_id, inode, blockno),
                )

                # Decrease the refcount, but don't take any further action
                # because the reference count might be wrong
                self.conn.execute(
                    'UPDATE objects SET refcount=refcount-1 WHERE id=?', (old_obj_id,)
                )
                self.unlinked_objects.add(old_obj_id)

            if not keep_cache:
                os.unlink(os.path.join(self.cachedir, filename))

    def check_lof(self):
        """Ensure that there is a lost+found directory"""

        log.info('Checking lost+found...')

        now_ns = time_ns()
        try:
            (inode_l, name_id) = self.conn.get_row(
                "SELECT inode, name_id FROM contents_v WHERE name=? AND parent_inode=?",
                (b"lost+found", ROOT_INODE),
            )

        except NoSuchRowError:
            self.found_errors = True
            try:
                # A bug in sqlite made fsck.s3ql move lost+found into lost+found. We just move it back to the root.
                (inode_l, name_id) = self.conn.get_row(
                    "SELECT inode, name_id FROM contents_v WHERE name=? AND parent_inode=inode",
                    (b"lost+found",),
                )
                self.log_error("Moving lost+found directory to root")
                self.conn.execute(
                    "UPDATE contents SET parent_inode = ? WHERE inode = ?", (ROOT_INODE, inode_l)
                )
            except NoSuchRowError:
                self.log_error("Recreating missing lost+found directory")
                inode_l = self.create_inode(
                    mode=stat.S_IFDIR | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR,
                    atime_ns=now_ns,
                    ctime_ns=now_ns,
                    mtime_ns=now_ns,
                    refcount=1,
                )
                self.conn.execute(
                    "INSERT INTO contents (name_id, inode, parent_inode) VALUES(?,?,?)",
                    (self._add_name(b"lost+found"), inode_l, ROOT_INODE),
                )

        mode = self.conn.get_val('SELECT mode FROM inodes WHERE id=?', (inode_l,))
        if not stat.S_ISDIR(mode):
            self.found_errors = True
            self.log_error(
                '/lost+found is not a directory! Old entry will be saved as '
                '/lost+found/inode-%s*',
                inode_l,
            )
            # We leave the old inode unassociated, so that it will be added
            # to lost+found later on.
            inode_l = self.create_inode(
                mode=stat.S_IFDIR | stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR,
                atime_ns=now_ns,
                ctime_ns=now_ns,
                mtime_ns=now_ns,
                refcount=1,
            )
            self.conn.execute(
                'UPDATE contents SET inode=? WHERE name_id=? AND parent_inode=?',
                (inode_l, name_id, ROOT_INODE),
            )

    def check_contents_name(self):
        """Check contents.name_id"""

        log.info('Checking contents (names)...')

        for rowid, name_id, inode_p, inode in self.conn.query(
            'SELECT contents.rowid, name_id, parent_inode, inode '
            'FROM contents LEFT JOIN names ON %(dbf)sname_id = %(dbf)snames.id '
            'WHERE %(dbf)snames.id IS NULL' % self.conn.fixes
        ):
            self.found_errors = True
            try:
                path = get_path(inode_p, self.conn)[1:]
            except NoSuchRowError:
                newname = ('-%d' % inode).encode()
            else:
                newname = escape(path) + ('-%d' % inode).encode()
            (id_p_new, newname) = self.resolve_free(b"/lost+found", newname)

            self.log_error(
                'Content entry for inode %d refers to non-existing name with id %d, '
                'moving to /lost+found/%s',
                inode,
                name_id,
                to_str(newname),
            )

            self.conn.execute(
                'UPDATE contents SET name_id=?, parent_inode=? WHERE rowid=?',
                (self._add_name(newname), id_p_new, rowid),
            )

    def check_contents_parent_inode(self):
        """Check contents.parent_inode"""

        log.info('Checking contents (parent inodes)...')

        for rowid, inode_p, name_id in self.conn.query(
            'SELECT contents.rowid, parent_inode, name_id '
            'FROM contents LEFT JOIN inodes '
            'ON %(dbf)sparent_inode = %(dbf)sinodes.id '
            'WHERE %(dbf)sinodes.id IS NULL' % self.conn.fixes
        ):
            self.found_errors = True
            name = self.conn.get_val('SELECT name FROM names WHERE id = ?', (name_id,))
            (id_p_new, newname) = self.resolve_free(
                b"/lost+found", ('[%d]-%s' % (inode_p, name)).encode()
            )

            self.log_error(
                'Parent inode %d for "%s" vanished, moving to /lost+found', inode_p, to_str(name)
            )
            self._del_name(name_id)
            self.conn.execute(
                'UPDATE contents SET name_id=?, parent_inode=? WHERE rowid=?',
                (self._add_name(newname), id_p_new, rowid),
            )

    def check_contents_inode(self):
        """Check contents.inode"""

        log.info('Checking contents (inodes)...')

        to_delete = list()
        for rowid, inode_p, inode, name_id in self.conn.query(
            'SELECT contents.rowid, parent_inode, inode, '
            'name_id FROM contents LEFT JOIN inodes '
            'ON %(dbf)sinode = %(dbf)sinodes.id '
            'WHERE %(dbf)sinodes.id IS NULL' % self.conn.fixes
        ):
            self.found_errors = True
            try:
                path = get_path(inode, self.conn)[1:]
            except NoSuchRowError:
                path = '[inode %d, parent %d]' % (inode, inode_p)

            self.log_error('Inode for %s vanished, deleting', to_str(path))
            self._del_name(name_id)
            to_delete.append(rowid)

        for rowid in to_delete:
            self.conn.execute('DELETE FROM contents WHERE rowid=?', (rowid,))

    def check_ext_attributes_name(self):
        """Check ext_attributes.name_id"""

        log.info('Checking extended attributes (names)...')

        for rowid, name_id, inode in self.conn.query(
            'SELECT ext_attributes.rowid, name_id, inode '
            'FROM ext_attributes LEFT JOIN names '
            'ON %(dbf)sname_id = %(dbf)snames.id '
            'WHERE %(dbf)snames.id IS NULL' % self.conn.fixes
        ):
            self.found_errors = True
            for name, id_p in self.conn.query(
                'SELECT name, parent_inode FROM contents_v WHERE inode=?', (inode,)
            ):
                path = get_path(id_p, self.conn, name)
                self.log_error(
                    'Extended attribute %d of %s refers to non-existing name %d, renaming..',
                    rowid,
                    to_str(path),
                    name_id,
                )

            while True:
                name_id = self._add_name('lost+found_%d' % rowid)
                if not self.conn.has_val(
                    "SELECT 1 FROM ext_attributes WHERE name_id=? AND inode=?", (name_id, inode)
                ):
                    self.conn.execute(
                        'UPDATE ext_attributes SET name_id=? WHERE rowid=?', (name_id, rowid)
                    )
                    break
                self._del_name('lost+found_%d' % rowid)
                rowid += 1

    def check_ext_attributes_inode(self):
        """Check ext_attributes.inode"""

        log.info('Checking extended attributes (inodes)...')

        to_delete = list()
        for rowid, inode, name_id in self.conn.query(
            'SELECT ext_attributes.rowid, inode, name_id '
            'FROM ext_attributes LEFT JOIN inodes '
            'ON %(dbf)sinode = %(dbf)sinodes.id '
            'WHERE %(dbf)sinodes.id IS NULL' % self.conn.fixes
        ):
            self.found_errors = True
            self.log_error(
                'Extended attribute %d refers to non-existing inode %d, deleting', rowid, inode
            )
            to_delete.append(rowid)
            self._del_name(name_id)

        for rowid in to_delete:
            self.conn.execute('DELETE FROM ext_attributes WHERE rowid=?', (rowid,))

    def check_loops(self):
        """Ensure that all directories can be reached from root"""

        log.info('Checking directory reachability...')

        self.conn.execute(
            'CREATE TEMPORARY TABLE loopcheck (inode INTEGER PRIMARY KEY, parent_inode INTEGER)'
        )
        self.conn.execute('CREATE INDEX ix_loopcheck_parent_inode ON loopcheck(parent_inode)')
        self.conn.execute(
            'INSERT INTO loopcheck (inode) '
            'SELECT parent_inode FROM contents GROUP BY parent_inode'
        )
        self.conn.execute(
            'UPDATE loopcheck SET parent_inode = '
            '(SELECT contents.parent_inode FROM contents '
            ' WHERE contents.inode = loopcheck.inode LIMIT 1)'
        )
        self.conn.execute('CREATE TEMPORARY TABLE loopcheck2 (inode INTEGER PRIMARY KEY)')
        self.conn.execute('INSERT INTO loopcheck2 (inode) SELECT inode FROM loopcheck')

        def delete_tree(inode_p):
            for (inode,) in self.conn.query(
                "SELECT inode FROM loopcheck WHERE parent_inode=?", (inode_p,)
            ):
                delete_tree(inode)
            self.conn.execute('DELETE FROM loopcheck2 WHERE inode=?', (inode_p,))

        root = ROOT_INODE
        while True:
            delete_tree(root)

            if not self.conn.has_val("SELECT 1 FROM loopcheck2"):
                break

            self.found_errors = True

            # Try obvious culprits first
            try:
                inode = self.conn.get_val(
                    'SELECT loopcheck2.inode FROM loopcheck2 JOIN contents '
                    'ON loopcheck2.inode = contents.inode '
                    'WHERE parent_inode = contents.inode LIMIT 1'
                )
            except NoSuchRowError:
                inode = self.conn.get_val("SELECT inode FROM loopcheck2 ORDER BY inode ASC LIMIT 1")

            (name, name_id) = self.conn.get_row(
                "SELECT name, name_id FROM contents_v WHERE inode=? LIMIT 1", (inode,)
            )
            (id_p, name) = self.resolve_free(b"/lost+found", name)

            self.log_error(
                "Found unreachable filesystem entries, re-anchoring %s [%d] in /lost+found",
                to_str(name),
                inode,
            )
            self.conn.execute(
                'UPDATE contents SET parent_inode=?, name_id=? WHERE inode=? AND name_id=?',
                (id_p, self._add_name(name), inode, name_id),
            )
            self._del_name(name_id)
            self.conn.execute('UPDATE loopcheck SET parent_inode=? WHERE inode=?', (id_p, inode))
            root = inode

        self.conn.execute("DROP TABLE loopcheck")
        self.conn.execute("DROP TABLE loopcheck2")

    def check_inodes_size(self):
        """Check inodes.size"""

        log.info('Checking inodes (sizes)...')

        self.conn.execute(
            'CREATE TEMPORARY TABLE min_sizes '
            '(id INTEGER PRIMARY KEY, min_size INTEGER NOT NULL)'
        )
        try:
            self.conn.execute(
                '''
            INSERT INTO min_sizes (id, min_size)
            SELECT inode, MAX(blockno * ? + length)
            FROM inode_blocks JOIN objects ON obj_id == objects.id
            GROUP BY inode''',
                (self.max_obj_size,),
            )

            self.conn.execute(
                '''
               CREATE TEMPORARY TABLE wrong_sizes AS
               SELECT id, size, min_size
                 FROM inodes JOIN min_sizes USING (id)
                WHERE size < min_size'''
            )

            for id_, size_old, size in self.conn.query('SELECT * FROM wrong_sizes'):
                self.found_errors = True
                self.log_error(
                    "Size of inode %d (%s) does not agree with number of blocks, "
                    "setting from %d to %d",
                    id_,
                    to_str(get_path(id_, self.conn)),
                    size_old,
                    size,
                )
                self.conn.execute("UPDATE inodes SET size=? WHERE id=?", (size, id_))
        finally:
            self.conn.execute('DROP TABLE min_sizes')
            self.conn.execute('DROP TABLE IF EXISTS wrong_sizes')

    def check_inodes_refcount(self):
        """Check inodes.refcount"""

        log.info('Checking inodes (refcounts)...')

        self.conn.execute(
            'CREATE TEMPORARY TABLE refcounts '
            '(id INTEGER PRIMARY KEY, refcount INTEGER NOT NULL)'
        )
        try:
            self.conn.execute(
                'INSERT INTO refcounts (id, refcount) '
                'SELECT inode, COUNT(name_id) FROM contents GROUP BY inode'
            )

            self.conn.execute(
                '''
               CREATE TEMPORARY TABLE wrong_refcounts AS
               SELECT inodes.id, refcounts.refcount, inodes.refcount
                 FROM inodes LEFT JOIN refcounts ON %(dbf)sinodes.id = %(dbf)srefcounts.id
                WHERE inodes.refcount != refcounts.refcount
                   OR %(dbf)srefcounts.refcount IS NULL'''
                % self.conn.fixes
            )

            for id_, cnt, cnt_old in self.conn.query('SELECT * FROM wrong_refcounts'):
                # No checks for root and control
                if id_ in (ROOT_INODE, CTRL_INODE):
                    continue

                self.found_errors = True
                if cnt is None:
                    (id_p, name) = self.resolve_free(b"/lost+found", ("inode-%d" % id_).encode())
                    self.log_error(
                        "Inode %d not referenced, adding as /lost+found/%s", id_, to_str(name)
                    )
                    self.conn.execute(
                        "INSERT INTO contents (name_id, inode, parent_inode) VALUES (?,?,?)",
                        (self._add_name(basename(name)), id_, id_p),
                    )
                    self.conn.execute("UPDATE inodes SET refcount=? WHERE id=?", (1, id_))

                else:
                    self.log_error(
                        "Inode %d (%s) has wrong reference count, setting from %d to %d",
                        id_,
                        to_str(get_path(id_, self.conn)),
                        cnt_old,
                        cnt,
                    )
                    self.conn.execute("UPDATE inodes SET refcount=? WHERE id=?", (cnt, id_))
        finally:
            self.conn.execute('DROP TABLE refcounts')
            self.conn.execute('DROP TABLE IF EXISTS wrong_refcounts')

    def check_inode_blocks_inode(self):
        """Check inode_blocks.inode"""

        log.info('Checking inode_blocks.inode...')

        to_delete = list()
        for rowid, inode, obj_id in self.conn.query(
            'SELECT inode_blocks.rowid, inode, obj_id '
            'FROM inode_blocks LEFT JOIN inodes '
            'ON %(dbf)sinode = %(dbf)sinodes.id '
            'WHERE %(dbf)sinodes.id IS NULL' % self.conn.fixes
        ):
            self.found_errors = True
            self.log_error(
                'Inode-block mapping %d refers to non-existing inode %d, deleting', rowid, inode
            )
            to_delete.append(rowid)

            # Decrease the refcount, but don't take any action because the
            # reference count might be wrong
            self.conn.execute('UPDATE objects SET refcount=refcount-1 WHERE id=?', (obj_id,))
            self.unlinked_objects.add(obj_id)

        for rowid in to_delete:
            self.conn.execute('DELETE FROM inode_blocks WHERE rowid=?', (rowid,))

    def check_inode_blocks_obj_id(self):
        """Check inode_blocks.obj_id"""

        log.info('Checking block-object mapping...')

        to_delete = list()
        for rowid, obj_id, inode in self.conn.query(
            'SELECT inode_blocks.rowid, obj_id, inode FROM inode_blocks '
            'LEFT JOIN objects ON %(dbf)sobj_id = %(dbf)sobjects.id '
            'WHERE %(dbf)sobjects.id IS NULL' % self.conn.fixes
        ):
            self.found_errors = True
            self.log_error(
                'Inode-block mapping for inode %d refers to non-existing object %d', inode, obj_id
            )
            to_delete.append(rowid)

            affected_entries = list(
                self.conn.query(
                    'SELECT name, name_id, parent_inode FROM contents_v WHERE inode=?', (inode,)
                )
            )
            for name, name_id, id_p in affected_entries:
                path = get_path(id_p, self.conn, name)
                self.log_error("File may lack data, moved to /lost+found: %s", to_str(path))
                (lof_id, newname) = self.resolve_free(b"/lost+found", escape(path))

                self.conn.execute(
                    'UPDATE contents SET name_id=?, parent_inode=? '
                    'WHERE name_id=? AND parent_inode=?',
                    (self._add_name(newname), lof_id, name_id, id_p),
                )
                self._del_name(name_id)

        for rowid in to_delete:
            self.conn.execute('DELETE FROM inode_blocks WHERE rowid=?', (rowid,))

    def check_symlinks_inode(self):
        """Check symlinks.inode"""

        log.info('Checking symlinks (inodes)...')

        to_delete = list()
        for rowid, inode in self.conn.query(
            'SELECT symlink_targets.rowid, inode FROM symlink_targets '
            'LEFT JOIN inodes ON %(dbf)sinode = %(dbf)sinodes.id '
            'WHERE %(dbf)sinodes.id IS NULL' % self.conn.fixes
        ):
            self.found_errors = True
            self.log_error('Symlink %d refers to non-existing inode %d, deleting', rowid, inode)
            to_delete.append(rowid)

        for rowid in to_delete:
            self.conn.execute('DELETE FROM symlink_targets WHERE rowid=?', (rowid,))

    def check_objects_hash(self):
        """Check objects.hash"""

        log.info('Checking objects.hash...')

        count = self.conn.get_val('SELECT COUNT(id) FROM objects WHERE hash IS NULL')
        if not count:
            return

        # This should only happen when there was an error during upload, so the
        # object must not have been stored correctly. We cannot just recalculate
        # the hash for the block, because then we may be modifying the contents
        # of the inode that refers to this block.
        self.found_errors = True
        self.log_error("No checksum for %d objects, removing from table...", count)
        self.conn.execute('DELETE FROM objects WHERE hash IS NULL')

        # Orphaned objects will be picked up by check_objects_id(), and missing
        # blocks handled by check_inode_blocks_obj_id().

    def create_inode(
        self,
        mode,
        uid=os.getuid(),
        gid=os.getgid(),
        mtime_ns=None,
        atime_ns=None,
        ctime_ns=None,
        refcount=None,
        size=0,
    ):
        '''Create inode'''

        id_ = self.conn.rowid(
            'INSERT INTO inodes (mode,uid,gid,mtime_ns,atime_ns,ctime_ns,'
            'refcount,size) VALUES (?,?,?,?,?,?,?,?)',
            (mode, uid, gid, mtime_ns, atime_ns, ctime_ns, refcount, size),
        )

        return id_

    def check_names_refcount(self):
        """Check names.refcount"""

        log.info('Checking names (refcounts)...')

        self.conn.execute(
            'CREATE TEMPORARY TABLE refcounts '
            '(id INTEGER PRIMARY KEY, refcount INTEGER NOT NULL)'
        )
        try:
            self.conn.execute('INSERT INTO refcounts (id, refcount) SELECT id, 0 FROM names')
            self.conn.execute(
                'UPDATE refcounts SET refcount='
                '(SELECT COUNT(name_id) FROM contents WHERE name_id = refcounts.id)'
                '+ (SELECT COUNT(name_id) FROM ext_attributes '
                '   WHERE name_id = refcounts.id)'
            )

            self.conn.execute(
                '''
               CREATE TEMPORARY TABLE wrong_refcounts AS
               SELECT names.id, refcounts.refcount, names.refcount
                 FROM names LEFT JOIN refcounts ON %(dbf)snames.id = %(dbf)srefcounts.id
                WHERE names.refcount != refcounts.refcount
                   OR %(dbf)srefcounts.refcount IS NULL'''
                % self.conn.fixes
            )

            for id_, cnt, cnt_old in self.conn.query('SELECT * FROM wrong_refcounts'):
                self.found_errors = True
                if cnt is None:
                    self.log_error(
                        "Name %d not referenced, removing (old refcount: %d)", id_, cnt_old
                    )
                    self.conn.execute('DELETE FROM names WHERE id=?', (id_,))
                else:
                    self.log_error(
                        "Name %d has wrong reference count, setting from %d to %d",
                        id_,
                        cnt_old,
                        cnt,
                    )
                    self.conn.execute("UPDATE names SET refcount=? WHERE id=?", (cnt, id_))
        finally:
            self.conn.execute('DROP TABLE refcounts')
            self.conn.execute('DROP TABLE IF EXISTS wrong_refcounts')

    def check_unix(self):
        """Check if file systems for agreement with UNIX conventions

        This means:
        - Only directories should have child entries
        - Only regular files should have data blocks and a size
        - Only symlinks should have a target
        - Only devices should have a device number
        - symlink size is length of target
        - names are not longer than 255 bytes
        - All directory entries have a valid mode

        Note that none of this is enforced by S3QL. However, as long as S3QL
        only communicates with the UNIX FUSE module, none of the above should
        happen (and if it does, it would probably confuse the system quite a
        lot).
        """

        log.info('Checking unix conventions...')

        for inode, mode, size, target, rdev in self.conn.query(
            "SELECT id, mode, size, target, rdev "
            "FROM inodes LEFT JOIN symlink_targets ON id = inode"
        ):
            has_children = self.conn.has_val(
                'SELECT 1 FROM contents WHERE parent_inode=? LIMIT 1', (inode,)
            )

            if stat.S_IFMT(mode) == 0:
                if has_children:
                    mode = mode | stat.S_IFDIR
                    made_to = 'directory'
                else:
                    mode = mode | stat.S_IFREG
                    made_to = 'regular file'

                self.found_errors = True
                self.log_error(
                    'Inode %d (%s): directory entry has no type, changed to %s.',
                    inode,
                    to_str(get_path(inode, self.conn)),
                    made_to,
                )
                self.conn.execute('UPDATE inodes SET mode=? WHERE id=?', (mode, inode))

            if stat.S_ISLNK(mode) and target is None:
                self.found_errors = True
                self.log_error(
                    'Inode %d (%s): symlink does not have target. '
                    'This is probably going to confuse your system!',
                    inode,
                    to_str(get_path(inode, self.conn)),
                )

            if stat.S_ISLNK(mode) and target is not None and size != len(target):
                self.found_errors = True
                self.log_error(
                    'Inode %d (%s): symlink size (%d) does not agree with target '
                    'length (%d). This is probably going to confuse your system!',
                    inode,
                    to_str(get_path(inode, self.conn)),
                    size,
                    len(target),
                )

            if size != 0 and (
                not stat.S_ISREG(mode) and not stat.S_ISLNK(mode) and not stat.S_ISDIR(mode)
            ):
                self.found_errors = True
                self.log_error(
                    'Inode %d (%s) is not regular file but has non-zero size. '
                    'This is may confuse your system!',
                    inode,
                    to_str(get_path(inode, self.conn)),
                )

            if target is not None and not stat.S_ISLNK(mode):
                self.found_errors = True
                self.log_error(
                    'Inode %d (%s) is not symlink but has symlink target. '
                    'This is probably going to confuse your system!',
                    inode,
                    to_str(get_path(inode, self.conn)),
                )

            if rdev != 0 and not (stat.S_ISBLK(mode) or stat.S_ISCHR(mode)):
                self.found_errors = True
                self.log_error(
                    'Inode %d (%s) is not device but has device number. '
                    'This is probably going to confuse your system!',
                    inode,
                    to_str(get_path(inode, self.conn)),
                )

            if has_children and not stat.S_ISDIR(mode):
                self.found_errors = True
                self.log_error(
                    'Inode %d (%s) is not a directory but has child entries. '
                    'This is probably going to confuse your system!',
                    inode,
                    to_str(get_path(inode, self.conn)),
                )

            if not stat.S_ISREG(mode) and self.conn.has_val(
                'SELECT 1 FROM inode_blocks WHERE inode=?', (inode,)
            ):
                self.found_errors = True
                self.log_error(
                    'Inode %d (%s) is not a regular file but has data blocks. '
                    'This is probably going to confuse your system!',
                    inode,
                    to_str(get_path(inode, self.conn)),
                )

        for name, id_p in self.conn.query(
            'SELECT name, parent_inode FROM contents_v WHERE LENGTH(name) > 255'
        ):
            path = get_path(id_p, self.conn, name)
            self.log_error(
                'Entry name %s... in %s has more than 255 characters, this could cause problems',
                to_str(name[:40]),
                to_str(path[: -len(name)]),
            )
            self.found_errors = True

    def check_objects_refcount(self):
        """Check objects.refcount"""

        log.info('Checking objects (reference counts)...')

        self.conn.execute(
            'CREATE TEMPORARY TABLE refcounts '
            '(id INTEGER PRIMARY KEY, refcount INTEGER NOT NULL)'
        )
        try:
            self.conn.execute(
                'INSERT INTO refcounts (id, refcount) '
                'SELECT obj_id, COUNT(obj_id) FROM inode_blocks GROUP BY obj_id'
            )

            self.conn.execute(
                '''
               CREATE TEMPORARY TABLE wrong_refcounts AS
               SELECT objects.id, refcounts.refcount, objects.refcount
                 FROM objects LEFT JOIN refcounts ON %(dbf)sobjects.id = %(dbf)srefcounts.id
                WHERE objects.refcount != refcounts.refcount
                   OR %(dbf)srefcounts.refcount IS NULL'''
                % self.conn.fixes
            )

            for id_, cnt, cnt_old in self.conn.query('SELECT * FROM wrong_refcounts'):
                if cnt is None and id_ in self.unlinked_objects and cnt_old == 0:
                    # Object was unlinked by check_cache
                    self.conn.execute('DELETE FROM objects WHERE id=?', (id_,))
                else:
                    self.found_errors = True
                    self.log_error(
                        "Object %s has invalid refcount, setting from %d to %d",
                        id_,
                        cnt_old,
                        cnt or 0,
                    )

                    if cnt is not None:
                        self.conn.execute("UPDATE objects SET refcount=? WHERE id=?", (cnt, id_))
                    else:
                        # Orphaned object will be picked up by check_objects_id(), dangling
                        # references by check_inode_blocks_obj_id()
                        self.conn.execute('DELETE FROM objects WHERE id=?', (id_,))
        finally:
            self.conn.execute('DROP TABLE refcounts')
            self.conn.execute('DROP TABLE IF EXISTS wrong_refcounts')

        # Delete objects which (correctly had) refcount=0
        for obj_id in self.conn.query('SELECT id FROM objects WHERE refcount = 0'):
            del self.backend['s3ql_data_%d' % obj_id]
        self.conn.execute("DELETE FROM objects WHERE refcount = 0")

    def check_objects_temp(self):
        """Remove temporary objects"""

        # Tests may provide a plain backend directly, but in regular operation
        # we'll always work with a ComprencBackend (even if there is neither
        # compression nor encryption)
        if isinstance(self.backend, ComprencBackend):
            plain_backend = self.backend.backend
        else:
            assert isinstance(self.backend, LocalBackend)
            plain_backend = self.backend

        if not isinstance(plain_backend, LocalBackend):
            return

        log.info('Checking for temporary objects and empty directories (backend)...')

        empty_dirs = 0
        for path, _, filenames in os.walk(plain_backend.prefix, topdown=False):
            for name in filenames:
                if not re.search(r'^[^#]+#[0-9]+--?[0-9]+\.tmp$', name):
                    continue

                self.found_errors = True
                self.log_error("removing temporary file %s", name)
                os.unlink(os.path.join(path, name))

            if path != plain_backend.prefix:
                try:
                    os.rmdir(path)
                    empty_dirs += 1
                except OSError:
                    pass

        log.info('Removed %d empty directories', empty_dirs)

    def check_objects_id(self):
        """Check objects.id"""

        log.info('Checking objects (backend)...')

        # We use this table to keep track of the objects that we have seen
        self.conn.execute("CREATE TEMP TABLE obj_ids (id INTEGER PRIMARY KEY)")
        try:
            for i, obj_name in enumerate(self.backend.list('s3ql_data_')):
                log.info(
                    'Processed %d objects so far',
                    i,
                    extra={'rate_limit': 1, 'update_console': True},
                )

                # We only bother with data objects
                try:
                    obj_id = int(obj_name[10:])
                except ValueError:
                    log.warning("Ignoring unexpected object %r", obj_name)
                    continue

                self.conn.execute('INSERT INTO obj_ids VALUES(?)', (obj_id,))

            for (obj_id,) in self.conn.query(
                'SELECT id FROM obj_ids EXCEPT SELECT id FROM objects'
            ):
                try:
                    if obj_id in self.unlinked_objects:
                        del self.backend['s3ql_data_%d' % obj_id]
                    else:
                        # TODO: Save the data in lost+found instead
                        del self.backend['s3ql_data_%d' % obj_id]
                        self.found_errors = True
                        self.log_error("Deleted spurious object %d", obj_id)
                except NoSuchObject:
                    pass

            self.conn.execute(
                'CREATE TEMPORARY TABLE missing AS '
                'SELECT id FROM objects EXCEPT SELECT id FROM obj_ids'
            )
            for (obj_id,) in self.conn.query('SELECT * FROM missing'):
                if ('s3ql_data_%d' % obj_id) in self.backend:
                    # Object was just not in list yet
                    continue

                self.found_errors = True
                self.log_error(
                    "object %s only exists in table but not in backend, deleting", obj_id
                )
                # Missing references will be picked up by check_inode_blocks_obj_id() later.
                self.conn.execute("DELETE FROM objects WHERE id=?", (obj_id,))

        finally:
            self.conn.execute('DROP TABLE obj_ids')
            self.conn.execute('DROP TABLE IF EXISTS missing')

    def check_objects_phys_size(self):
        """Check objects.phys_size"""

        log.info('Checking objects (physical sizes)...')

        for (obj_id,) in self.conn.query('SELECT id FROM objects WHERE phys_size = -1'):
            self.found_errors = True
            self.log_error("Object %d has no size information, retrieving from backend...", obj_id)
            self.conn.execute(
                'UPDATE objects SET phys_size=? WHERE id=?',
                (self.backend.get_size('s3ql_data_%d' % obj_id), obj_id),
            )

    def resolve_free(self, path, name):
        '''Return parent inode and name of an unused directory entry

        The directory entry will be in `path`. If an entry `name` already
        exists there, we append a numeric suffix.
        '''

        if not isinstance(path, bytes):
            raise TypeError('path must be of type bytes')
        if not isinstance(name, bytes):
            raise TypeError('name must be of type bytes')

        inode_p = inode_for_path(path, self.conn)

        # Debugging http://code.google.com/p/s3ql/issues/detail?id=217
        # and http://code.google.com/p/s3ql/issues/detail?id=261
        if len(name) > 255 - 4:
            name = b''.join((name[0:120], b' ... ', name[-120:]))

        i = 0
        newname = name
        name += b'-'
        try:
            while True:
                self.conn.get_val(
                    "SELECT inode FROM contents_v WHERE name=? AND parent_inode=?",
                    (newname, inode_p),
                )
                i += 1
                newname = name + str(i).encode()

        except NoSuchRowError:
            pass

        return (inode_p, newname)

    def _add_name(self, name):
        '''Get id for *name* and increase refcount

        Name is inserted in table if it does not yet exist.
        '''

        try:
            name_id = self.conn.get_val('SELECT id FROM names WHERE name=?', (name,))
        except NoSuchRowError:
            name_id = self.conn.rowid('INSERT INTO names (name, refcount) VALUES(?,?)', (name, 1))
        else:
            self.conn.execute('UPDATE names SET refcount=refcount+1 WHERE id=?', (name_id,))
        return name_id

    def _del_name(self, name_id):
        '''Decrease refcount for name_id, remove if it reaches 0'''

        self.conn.execute('UPDATE names SET refcount=refcount-1 WHERE id=?', (name_id,))
        self.conn.execute('DELETE FROM names WHERE refcount=0 AND id=?', (name_id,))


def parse_args(args):
    parser = ArgumentParser(description="Checks and repairs an S3QL filesystem.")

    parser.add_log('~/.s3ql/fsck.log')
    parser.add_cachedir()
    parser.add_debug()
    parser.add_quiet()
    parser.add_backend_options()
    parser.add_version()
    parser.add_storage_url()
    parser.add_compress()

    parser.add_argument(
        "--keep-cache",
        action="store_true",
        default=False,
        help="Do not purge locally cached files on exit.",
    )
    parser.add_argument(
        "--batch",
        action="store_true",
        default=False,
        help="If user input is required, exit without prompting.",
    )
    parser.add_argument(
        "--force",
        action="store_true",
        default=False,
        help="Force checking even if file system is marked clean.",
    )
    parser.add_argument(
        "--force-remote",
        action="store_true",
        default=False,
        help="Force use of remote metadata even when this would likely result in data loss.",
    )
    options = parser.parse_args(args)

    return options


def main(args=None):
    if args is None:
        args = sys.argv[1:]

    setup_warnings()
    options = parse_args(args)
    setup_logging(options)

    # Check if fs is mounted on this computer
    # This is not foolproof but should prevent common mistakes
    if is_mounted(options.storage_url):
        raise QuietError('Can not check mounted file system.', exitcode=40)

    backend = get_backend(options)
    atexit.register(backend.close)

    log.info('Starting fsck of %s', options.storage_url)

    cachepath = options.cachepath
    db = None

    param = read_remote_params(backend)
    local_param = read_cached_params(cachepath, min_seq=param.seq_no)
    if local_param is not None:
        log.info('Using cached metadata.')
        param = local_param
        db = Connection(cachepath + '.db', param.metadata_block_size)
        if param.is_mounted:
            log.info('File system was not unmounted cleanly')
            param.needs_fsck = True
        if local_param.seq_no > param.seq_no:
            log.info('Remote metadata is outdated.')
            param.needs_fsck = True

    if param.is_mounted and local_param is not param:
        print(
            textwrap.fill(
                textwrap.dedent(
                    '''\
              Backend reports that file system is still mounted elsewhere. If this is not
              true,  run fsck on the computer where the file system has been mounted most
              recently.

              You may also continue and use whatever metadata is available in the backend.
              However, in that case YOU MAY LOOSE ALL DATA THAT HAS BEEN UPLOADED OR MODIFIED
              SINCE THE LAST SUCCESSFUL METADATA UPLOAD. Moreover, files and directories that
              you have deleted since then MAY REAPPEAR WITH SOME OF THEIR CONTENT LOST.
              '''
                )
            )
        )

        print(
            'Enter "continue, I know what I am doing" to use the outdated data anyway:',
            '> ',
            sep='\n',
            end='',
        )
        if options.force_remote:
            print('(--force-remote specified, continuing anyway)')
        elif options.batch:
            raise QuietError('(in batch mode, exiting)', exitcode=41)
        elif sys.stdin.readline().strip() != 'continue, I know what I am doing':
            raise QuietError(exitcode=42)
        param.needs_fsck = True

    if (
        not param.needs_fsck and (time.time() - param.last_fsck) < 60 * 60 * 24 * 31
    ):  # last check more than 1 month ago
        if options.force:
            log.info('File system seems clean, checking anyway.')
        else:
            log.info('File system is marked as clean. Use --force to force checking.')
            return

    # When using remote metadata, get rid of outdated local cache (so that we
    # don't accidentally upload it)
    outdated_cachedir = False
    if not db:
        try:
            for name_ in os.listdir(cachepath + '-cache'):
                if name_ not in ('.', '..'):
                    outdated_cachedir = True
                    break
        except FileNotFoundError:
            pass
    if outdated_cachedir and param.needs_fsck:
        for i in itertools.count():
            bak_name = '%s-cache.bak%d' % (cachepath, i)
            if not os.path.exists(bak_name):
                break
        log.warning('Renaming outdated cache directory %s to .bak%d', cachepath + '-cache', i)
        log.warning(
            'You should delete this directory once you are sure that everything is in order.'
        )
        os.rename(cachepath + '-cache', bak_name)
    elif outdated_cachedir:
        log.info("Flushing outdated local cache...")
        shutil.rmtree(cachepath + '-cache')

    # We only read cache files if the filesystem was not unmounted cleanly. On a clean unmount, the
    # cache files can not be dirty.
    if db and param.is_mounted:
        if options.keep_cache:
            check_cache = 'keep'
        else:
            check_cache = True
    else:
        check_cache = False

    if not db:
        # When file system was not unmounted cleanly, the checksum and size may not have been
        # updated.
        log.info('Downloading metadata...')
        db = download_metadata(backend, cachepath + '.db', param, failsafe=param.is_mounted)

        # Don't download the same metadata again in verify_metadata_snapshots
        check_current_metadata = False
    else:
        # If the filesystem was not unmounted cleanly, do not bother to check the
        # current remote metadata (it's most likely inconsistent).
        check_current_metadata = not param.is_mounted

    log.info('Checking DB integrity...')
    try:
        # get_list may raise CorruptError itself
        res = db.get_list('PRAGMA integrity_check(20)')
        if res[0][0] != 'ok':
            log.error('\n'.join(x[0] for x in res))
            raise apsw.CorruptError()
    except apsw.CorruptError:
        raise QuietError(
            'Database file (f{cachepath}.db) is corrupted. Restore from a backup or try\n'
            'to repair with the SQLite CLI (cf. https://www.sqlite.org/recovery.html),\n'
            'then re-run fsck.s3ql',
            exitcode=43,
        )

    # To detect bugs in S3QL, make sure we can correctly download the last few metadata snapshots.
    verify_metadata_snapshots(backend, count=5, include_most_recent=check_current_metadata)

    param.is_mounted = True
    param.seq_no += 1
    write_params(cachepath, param)
    upload_params(backend, param)

    fsck = Fsck(cachepath + '-cache', backend, param, db)
    fsck.check(check_cache)

    if fsck.uncorrectable_errors:
        raise QuietError("Uncorrectable errors found, aborting.", exitcode=44 + 128)

    if fsck.found_errors and not param.needs_fsck:
        log.error('File system was marked as clean, yet fsck found problems.')
        log.error(
            'Please report this to the S3QL mailing list, http://groups.google.com/group/s3ql'
        )

    # If the filesystem was not unmounted cleanly, the database may have changes that have not
    # been uploaded yet are not tracked through the VFS, so we need to do a full upload. Just
    # to be sure, we also do a full upload if we found any errors.
    if param.needs_fsck or fsck.found_errors:
        full_upload = True
    else:
        full_upload = False

    param.needs_fsck = False
    param.is_mounted = False
    param.last_fsck = time.time()
    param.last_modified = time.time()

    if full_upload:
        # This is a good time to reduce DB size if possible
        db.execute('VACUUM')
        db.close()
        upload_metadata(backend, db, param, incremental=False)
    else:
        db.close()
        upload_metadata(backend, db, param, incremental=True)

    write_params(cachepath, param)
    upload_params(backend, param)

    expire_objects(backend)

    log.info('Completed fsck of %s', options.storage_url)

    if fsck.found_errors:
        sys.exit(128)
    else:
        sys.exit(0)


def escape(path):
    '''Escape slashes in path so that is usable as a file name'''

    return path[1:].replace(b'_', b'__').replace(b'/', b'_')


def to_str(name):
    '''Decode path name for printing'''

    return str(name, encoding='utf-8', errors='replace')


def verify_metadata_snapshots(backend, count: int = 5, include_most_recent: bool = True):
    '''Verify the last *count* metadata snapshots.'''

    log.info('Verifying consistency of most recent metadata backups:')
    backups = sorted(get_available_seq_nos(backend), reverse=True)

    if not backups:
        raise RuntimeError(
            'No metadata backups found. This should not happen, please report a bug.'
        )

    if include_most_recent:
        to_check = backups[:count]
    else:
        to_check = backups[1 : count + 1]

    for seq_no in to_check:
        # De-serialize directly instead of using read_remote_params() to avoid
        # exceptions on old filesystem revisions.
        d = thaw_basic_mapping(backend['s3ql_params_%010x' % seq_no])
        if d['revision'] != CURRENT_FS_REV:
            break
        params = FsAttributes(**d)
        assert params.seq_no == seq_no
        date = datetime.fromtimestamp(params.last_modified).strftime('%Y-%m-%d %H:%M:%S')
        if params.is_mounted:
            log.info(
                'Skipping check of backup %010x (from %s): not unmounted cleanly.', seq_no, date
            )
            continue
        log.info('Checking backup %010x (from %s)...', seq_no, date)
        with NamedTemporaryFile() as fh:
            conn = download_metadata(backend, fh.name, params)
            res = conn.get_list('PRAGMA integrity_check(20)')
            if res[0][0] != 'ok':
                log.error('\n'.join(x[0] for x in res))
                raise RuntimeError(
                    'Metadata backup is corrupted. Please report this as a bug',
                )
            conn.close()


if __name__ == '__main__':
    main(sys.argv[1:])



from ..s3c import HTTPError


class B2Error(HTTPError):
    '''
    Represents an error returned by Backblaze B2 API call

    For possible codes, see https://www.backblaze.com/b2/docs/calling.html
    '''

    def __init__(self, status, code, message, headers=None):
        super().__init__(status, message, headers)
        self.code = code

        # Force 1s waiting time before retry
        if not self.retry_after:
            self.retry_after = 1

    def __str__(self):
        return '%s : %s - %s' % (self.status, self.code, self.msg)


class BadDigestError(B2Error):
    pass



'''
fs.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import collections
import errno
import logging
import math
import os
import stat
import struct
import time
from io import BytesIO

import pyfuse3
import trio
from pyfuse3 import FUSEError

from . import CTRL_INODE, CTRL_NAME
from .backends.common import CorruptedObjectError, NoSuchObject
from .common import get_path, parse_literal, time_ns
from .database import Connection, NoSuchRowError

# We work in bytes
CTRL_NAME = CTRL_NAME.encode('us-ascii')

# standard logger for this module
log = logging.getLogger(__name__)

# During long-running operations, yield to event loop with at least
# this interval.
CHECKPOINT_INTERVAL = 0.05

# ACL_ERRNO is the error code returned for requests that try
# to modify or access extendeda attributes associated with ACL.
# Since we currently don't know how to keep these in sync
# with permission bits, we cannot support ACLs despite having
# full support for extended attributes.
#
# What errno to use for ACL_ERRNO is a bit tricky. acl_set_fd(3)
# returns ENOTSUP if the file system does not support ACLs. However,
# this function uses the setxattr() syscall internally, and
# setxattr(3) states that ENOTSUP means that the file system does not
# support extended attributes at all. A test with btrfs mounted with
# -o noacl shows that the actual errno returned by setxattr() is
# EOPNOTSUPP. Also, some Python versions do not know about
# errno.ENOTSUPP and errno(3) says that on Linux, EOPNOTSUPP and ENOTSUP
# have the same value (despite this violating POSIX).
#
# All in all, the situation seems complicated, so we try to use
# EOPNOTSUPP with a fallback on ENOTSUP just in case.

if not hasattr(errno, 'EOPNOTSUPP'):
    ACL_ERRNO = errno.ENOTSUP
else:
    ACL_ERRNO = errno.EOPNOTSUPP


class _TruncSetattrFields:
    """A private class used for O_TRUNC.

    This is needed because pyfuse3.SetattrFields fields are read only"""

    def __getattr__(self, name):
        """Only return ``True`` for ``update_size`` and ``update_mtime`.
        Anything else will return ``False``"""
        return name in ('update_size', 'update_mtime')


"""We only need a single instance of this read-only class"""
_TruncSetattrFields = _TruncSetattrFields()


class Operations(pyfuse3.Operations):
    """A full-featured file system for online data storage

    This class implements low-level FUSE operations and is meant to be passed to
    pyfuse3.init().

    The ``access`` method of this class always gives full access, independent of
    file permissions. If the FUSE library is initialized with ``allow_other`` or
    ``allow_root``, the ``default_permissions`` option should therefore always
    be passed as well.


    Attributes:
    -----------

    :cache:       Holds information about cached blocks
    :inode_cache: A cache for the attributes of the currently opened inodes.
    :open_inodes: dict of currently opened inodes. This is used to not remove
                  the blocks of unlinked inodes that are still open.
    :upload_task: Trio task for metadata uploads
    :failsafe: Set when backend problems are encountered. In that case, fs only
               allows read access.
    :broken_blocks: Caches information about corrupted blocks to avoid repeated (pointless)
                    attempts to retrieve them. This attribute is a dict (indexed by inodes)
                    of sets of block indices. Broken blocks are removed from the cache
                    when an inode is forgotten.

    Directory Entry Types
    ----------------------

    S3QL is quite agnostic when it comes to directory entry types. Every
    directory entry can contain other entries *and* have a associated data,
    size, link target and device number. However, S3QL makes some provisions for
    users relying on unlink()/rmdir() to fail for a directory/file. For that, it
    explicitly checks the st_mode attribute.
    """

    supports_dot_lookup = True
    enable_acl = False
    enable_writeback_cache = True

    def __init__(self, block_cache, db: Connection, max_obj_size, inode_cache, upload_task=None):
        super().__init__()

        self.inodes = inode_cache
        self.db = db
        self.upload_task = upload_task
        self.open_inodes = collections.defaultdict(lambda: 0)
        self.max_obj_size = max_obj_size
        self.cache = block_cache
        self.failsafe = False
        self.broken_blocks = collections.defaultdict(set)

        # Root inode is always open
        self.open_inodes[pyfuse3.ROOT_INODE] += 1

    async def destroy(self):
        await self.forget(list(self.open_inodes.items()))
        self.inodes.destroy()

    async def lookup(self, id_p, name, ctx):
        return self._lookup(id_p, name, ctx).entry_attributes()

    def _lookup(self, id_p, name, ctx):
        log.debug('started with %d, %r', id_p, name)

        if name == CTRL_NAME:
            inode = self.inodes[CTRL_INODE]

            # Make sure the control file is only writable by the user
            # who mounted the file system (but don't mark inode as dirty)
            object.__setattr__(inode, 'uid', os.getuid())
            object.__setattr__(inode, 'gid', os.getgid())

        elif name == '.':
            inode = self.inodes[id_p]

        elif name == '..':
            id_ = self.db.get_val("SELECT parent_inode FROM contents WHERE inode=?", (id_p,))
            inode = self.inodes[id_]

        else:
            try:
                id_ = self.db.get_val(
                    "SELECT inode FROM contents_v WHERE name=? AND parent_inode=?", (name, id_p)
                )
            except NoSuchRowError:
                raise FUSEError(errno.ENOENT)
            inode = self.inodes[id_]

        self.open_inodes[inode.id] += 1
        return inode

    async def getattr(self, id_, ctx):
        log.debug('started with %d', id_)
        if id_ == CTRL_INODE:
            # Make sure the control file is only writable by the user
            # who mounted the file system (but don't mark inode as dirty)
            inode = self.inodes[CTRL_INODE]
            object.__setattr__(inode, 'uid', os.getuid())
            object.__setattr__(inode, 'gid', os.getgid())
            return inode.entry_attributes()

        return self.inodes[id_].entry_attributes()

    async def readlink(self, id_, ctx):
        log.debug('started with %d', id_)
        now_ns = time_ns()
        inode = self.inodes[id_]
        if inode.atime_ns < inode.ctime_ns or inode.atime_ns < inode.mtime_ns:
            inode.atime_ns = now_ns
        try:
            return self.db.get_val("SELECT target FROM symlink_targets WHERE inode=?", (id_,))
        except NoSuchRowError:
            log.warning('Inode does not have symlink target: %d', id_)
            raise FUSEError(errno.EINVAL)

    async def opendir(self, id_, ctx):
        log.debug('started with %d', id_)
        return id_

    async def readdir(self, id_, off, token):
        log.debug('started with %d, %d', id_, off)
        if off == 0:
            off = -1

        inode = self.inodes[id_]
        if inode.atime_ns < inode.ctime_ns or inode.atime_ns < inode.mtime_ns:
            inode.atime_ns = time_ns()

        # NFS treats offsets 1 and 2 special, so we have to exclude
        # them.
        with self.db.query(
            "SELECT name_id, name, inode FROM contents_v "
            'WHERE parent_inode=? AND name_id > ? ORDER BY name_id',
            (id_, off - 3),
        ) as res:
            for next_, name, cid_ in res:
                if not pyfuse3.readdir_reply(
                    token, name, self.inodes[cid_].entry_attributes(), next_ + 3
                ):
                    break
                self.open_inodes[cid_] += 1

    async def getxattr(self, id_, name, ctx):
        log.debug('started with %d, %r', id_, name)
        # Handle S3QL commands
        if id_ == CTRL_INODE:
            if name == b's3ql_pid?':
                return ('%d' % os.getpid()).encode()

            elif name == b's3qlstat':
                return self.extstat()

            raise FUSEError(errno.EINVAL)

        # http://code.google.com/p/s3ql/issues/detail?id=385
        elif name in (b'system.posix_acl_access', b'system.posix_acl_default'):
            raise FUSEError(ACL_ERRNO)

        else:
            try:
                value = self.db.get_val(
                    'SELECT value FROM ext_attributes_v WHERE inode=? AND name=?', (id_, name)
                )
            except NoSuchRowError:
                raise FUSEError(pyfuse3.ENOATTR)
            return value

    async def listxattr(self, id_, ctx):
        log.debug('started with %d', id_)
        names = list()
        for (name,) in self.db.query('SELECT name FROM ext_attributes_v WHERE inode=?', (id_,)):
            names.append(name)
        return names

    async def setxattr(self, id_, name, value, ctx):
        log.debug('started with %d, %r, %r', id_, name, value)

        # Handle S3QL commands
        if id_ == CTRL_INODE:
            if name == b's3ql_flushcache!':
                self.inodes.flush()
                await self.cache.flush()

            elif name == b's3ql_dropcache!':
                self.inodes.drop()
                await self.cache.drop()

            elif name == b'copy':
                try:
                    tup = parse_literal(value, (int, int))
                except ValueError:
                    log.warning('Received malformed command via control inode')
                    raise FUSEError.EINVAL()
                await self.copy_tree(*tup)

            elif name == b'upload-meta':
                if self.upload_task is not None:
                    self.inodes.flush()
                    self.upload_task.event.set()
                else:
                    raise FUSEError(errno.ENOTTY)

            elif name == b'lock':
                try:
                    id_ = parse_literal(value, int)
                except ValueError:
                    log.warning('Received malformed command via control inode')
                    raise FUSEError.EINVAL()
                await self.lock_tree(id_)

            elif name == b'rmtree':
                try:
                    tup = parse_literal(value, (int, bytes))
                except ValueError:
                    log.warning('Received malformed command via control inode')
                    raise FUSEError.EINVAL()
                await self.remove_tree(*tup)

            elif name == b'logging':
                try:
                    (lvl, modules) = parse_literal(value, (int, str))
                except (ValueError, KeyError):
                    log.warning('Received malformed command via control inode')
                    raise FUSEError.EINVAL()
                update_logging(lvl, modules.split(',') if modules else None)

            elif name == b'cachesize':
                try:
                    self.cache.cache.max_size = parse_literal(value, int)
                except ValueError:
                    log.warning('Received malformed command via control inode')
                    raise FUSEError.EINVAL()
                log.debug('updated cache size to %d bytes', self.cache.cache.max_size)

            else:
                log.warning('Received unknown command via control inode')
                raise FUSEError(errno.EINVAL)

        # http://code.google.com/p/s3ql/issues/detail?id=385
        elif name in (b'system.posix_acl_access', b'system.posix_acl_default'):
            raise FUSEError(ACL_ERRNO)

        else:
            if self.failsafe or self.inodes[id_].locked:
                raise FUSEError(errno.EPERM)

            self.db.execute(
                'INSERT OR REPLACE INTO ext_attributes (inode, name_id, value) VALUES(?, ?, ?)',
                (id_, self._add_name(name), value),
            )
            self.inodes[id_].ctime_ns = time_ns()

    async def removexattr(self, id_, name, ctx):
        log.debug('started with %d, %r', id_, name)

        if self.failsafe or self.inodes[id_].locked:
            raise FUSEError(errno.EPERM)

        try:
            name_id = self._del_name(name)
        except NoSuchRowError:
            raise FUSEError(pyfuse3.ENOATTR)

        changes = self.db.execute(
            'DELETE FROM ext_attributes WHERE inode=? AND name_id=?', (id_, name_id)
        )
        if changes == 0:
            raise FUSEError(pyfuse3.ENOATTR)

        self.inodes[id_].ctime_ns = time_ns()

    async def lock_tree(self, id0):
        '''Lock directory tree'''

        if self.failsafe:
            raise FUSEError(errno.EPERM)

        log.debug('started with %d', id0)
        queue = [(id0, -1)]
        self.inodes[id0].locked = True
        conn = self.db

        # Batching updates makes things much faster when using WAL. We don't want to rollback on
        # error, because in that case the database gets out of sync with the inode cache (which
        # probably makes things worse).
        processed = 0
        with conn.batch(CHECKPOINT_INTERVAL) as batch_mgr:
            while queue:
                (id_p, off) = queue.pop()
                log.debug('Processing directory (%d, %d)', id_p, off)

                with conn.query(
                    'SELECT name_id, inode FROM contents WHERE parent_inode=? '
                    'AND name_id > ? ORDER BY name_id',
                    (id_p, off),
                ) as res:
                    for name_id, id_ in res:
                        self.inodes[id_].locked = True

                        if conn.has_val('SELECT 1 FROM contents WHERE parent_inode=?', (id_,)):
                            queue.append((id_, -1))

                        # Break every once in a while - note that we can't yield
                        # right here because there's an active DB query.
                        processed += 1
                        if processed > batch_mgr.batch_size:
                            queue.append((id_p, name_id))
                            break

                if processed > batch_mgr.batch_size:
                    batch_mgr.finish_batch(processed)
                    await trio.lowlevel.checkpoint()
                    processed = 0
                    batch_mgr.start_batch()

        log.debug('finished')

    async def remove_tree(self, id_p0, name0):
        '''Remove directory tree'''

        if self.failsafe:
            raise FUSEError(errno.EPERM)

        log.debug('started with %d, %s', id_p0, name0)

        if self.inodes[id_p0].locked:
            raise FUSEError(errno.EPERM)

        conn = self.db
        id0 = self._lookup(id_p0, name0, ctx=None).id
        queue = [id0]  # Directories that we still need to delete

        # Batching updates makes things much faster when using WAL. We don't want to rollback on
        # error, because in that case the database gets out of sync with the inode cache (which
        # probably makes things worse).
        processed = 0
        with conn.batch(CHECKPOINT_INTERVAL) as batch_mgr:
            while queue:  # For every directory
                id_p = queue.pop()
                is_open = id_p in self.open_inodes

                # Per https://sqlite.org/isolation.html, results of removing rows
                # during select are undefined. Therefore, process data in chunks.
                # This is also a nice opportunity to release the GIL...
                query_chunk = conn.get_list(
                    'SELECT name, name_id, inode FROM contents_v WHERE '
                    'parent_inode=? LIMIT %d' % batch_mgr.batch_size,
                    (id_p,),
                )
                reinserted = False

                for name, _, id_ in query_chunk:
                    if conn.has_val('SELECT 1 FROM contents WHERE parent_inode=?', (id_,)):
                        # First delete subdirectories
                        if not reinserted:
                            queue.append(id_p)
                            reinserted = True
                        queue.append(id_)
                    else:
                        if is_open:
                            # This may fail with ENOTEMPTY in rare circumstances. See below for details.
                            pyfuse3.invalidate_entry_async(
                                id_p, name, deleted=id_, ignore_enoent=True
                            )
                        await self._remove(id_p, name, id_, force=True)
                    processed += 1

                if len(query_chunk) == batch_mgr.batch_size and not reinserted:
                    # Make sure to re-insert the directory to process the remaining
                    # contents and delete the directory itself.
                    queue.append(id_p)

                if processed > batch_mgr.batch_size:
                    batch_mgr.finish_batch(processed)
                    await trio.lowlevel.checkpoint()
                    processed = 0
                    batch_mgr.start_batch()

        if id_p0 in self.open_inodes:
            log.debug('invalidate_entry(%d, %r)', id_p0, name0)
            # This may fail with ENOTEMPTY if the control file has been looked up underneath this
            # path (same for the invalidate_entry_call_async above). In normal operation this should
            # be rare, since the control file is not included in readdir() output and only S3QL
            # tools know about its existence. We could preemtively run invalidate_entry_async() for
            # the control file whenever we run it for a directory, but that would still leave a
            # window in which the file could be looked up again and would also impose a performance
            # penalty for something that is exceedingly rare. Tracked in https://github.com/s3ql/s3ql/issues/319.
            pyfuse3.invalidate_entry_async(id_p0, name0, deleted=id0, ignore_enoent=True)
        await self._remove(id_p0, name0, id0, force=True)

        await self.forget([(id0, 1)])
        log.debug('finished')

    async def copy_tree(self, src_id, target_id):
        '''Efficiently copy directory tree'''

        if self.failsafe:
            raise FUSEError(errno.EPERM)

        log.debug('started with %d, %d', src_id, target_id)

        # To avoid lookups and make code tidier
        make_inode = self.inodes.create_inode
        db = self.db

        # Copy target attributes
        # These come from setxattr, so they may have been deleted
        # without being in open_inodes
        try:
            src_inode = self.inodes[src_id]
            target_inode = self.inodes[target_id]
        except KeyError:
            raise FUSEError(errno.ENOENT)
        for attr in ('atime_ns', 'ctime_ns', 'mtime_ns', 'mode', 'uid', 'gid'):
            setattr(target_inode, attr, getattr(src_inode, attr))

        # We first replicate into a dummy inode, so that we
        # need to invalidate only once.
        now_ns = time_ns()
        tmp = make_inode(
            mtime_ns=now_ns, ctime_ns=now_ns, atime_ns=now_ns, uid=0, gid=0, mode=0, refcount=0
        )

        queue = [(src_id, tmp.id, -1)]
        id_cache = dict()

        # Batching updates makes things much faster when using WAL. We don't want to rollback on
        # error, because in that case the database gets out of sync with the inode cache (which
        # probably makes things worse).
        processed = 0
        with db.batch(CHECKPOINT_INTERVAL) as batch_mgr:
            while queue:
                (src_id, target_id, off) = queue.pop()
                # log.debug('Processing directory (%d, %d, %d)', src_id, target_id, off)

                with db.query(
                    'SELECT name_id, inode FROM contents WHERE parent_inode=? '
                    'AND name_id > ? ORDER BY name_id',
                    (src_id, off),
                ) as res:
                    for name_id, id_ in res:
                        # Make sure that all blocks are in the database
                        if id_ in self.open_inodes:
                            await self.cache.start_flush(id_)

                        if id_ not in id_cache:
                            inode = self.inodes[id_]
                            inode_new = make_inode(
                                refcount=1,
                                mode=inode.mode,
                                size=inode.size,
                                uid=inode.uid,
                                gid=inode.gid,
                                mtime_ns=inode.mtime_ns,
                                atime_ns=inode.atime_ns,
                                ctime_ns=inode.ctime_ns,
                                rdev=inode.rdev,
                            )

                            id_new = inode_new.id

                            if inode.refcount != 1:
                                id_cache[id_] = id_new

                            db.execute(
                                'INSERT INTO symlink_targets (inode, target) '
                                'SELECT ?, target FROM symlink_targets WHERE inode=?',
                                (id_new, id_),
                            )

                            db.execute(
                                'INSERT INTO ext_attributes (inode, name_id, value) '
                                'SELECT ?, name_id, value FROM ext_attributes WHERE inode=?',
                                (id_new, id_),
                            )
                            db.execute(
                                'UPDATE names SET refcount = refcount + 1 WHERE '
                                'id IN (SELECT name_id FROM ext_attributes WHERE inode=?)',
                                (id_,),
                            )

                            processed += db.execute(
                                'INSERT INTO inode_blocks (inode, blockno, obj_id) '
                                'SELECT ?, blockno, obj_id FROM inode_blocks '
                                'WHERE inode=?',
                                (id_new, id_),
                            )
                            db.execute(
                                'REPLACE INTO objects (id, hash, refcount, phys_size, length) '
                                'SELECT id, hash, refcount+COUNT(id), phys_size, length '
                                'FROM inode_blocks JOIN objects ON obj_id = id '
                                'WHERE inode = ? GROUP BY id',
                                (id_new,),
                            )

                            if db.has_val('SELECT 1 FROM contents WHERE parent_inode=?', (id_,)):
                                queue.append((id_, id_new, -1))
                        else:
                            id_new = id_cache[id_]
                            self.inodes[id_new].refcount += 1

                        db.execute(
                            'INSERT INTO contents (name_id, inode, parent_inode) VALUES(?, ?, ?)',
                            (name_id, id_new, target_id),
                        )
                        db.execute('UPDATE names SET refcount=refcount+1 WHERE id=?', (name_id,))

                        # Can't finish batch here because there's an active DB query.
                        processed += 1
                        if processed >= batch_mgr.batch_size:
                            queue.append((src_id, target_id, name_id))
                            break

                if processed >= batch_mgr.batch_size:
                    batch_mgr.finish_batch(processed)
                    await trio.lowlevel.checkpoint()
                    processed = 0
                    batch_mgr.start_batch()

        # Make replication visible
        self.db.execute(
            'UPDATE contents SET parent_inode=? WHERE parent_inode=?', (target_inode.id, tmp.id)
        )
        del self.inodes[tmp.id]
        pyfuse3.invalidate_inode(target_inode.id)

        log.debug('finished')

    async def unlink(self, id_p, name, ctx):
        log.debug('started with %d, %r', id_p, name)
        if self.failsafe:
            raise FUSEError(errno.EPERM)

        inode = self._lookup(id_p, name, ctx)

        if stat.S_ISDIR(inode.mode):
            raise FUSEError(errno.EISDIR)

        await self._remove(id_p, name, inode.id)

        await self.forget([(inode.id, 1)])

    async def rmdir(self, id_p, name, ctx):
        log.debug('started with %d, %r', id_p, name)
        if self.failsafe:
            raise FUSEError(errno.EPERM)

        inode = self._lookup(id_p, name, ctx)

        if self.inodes[id_p].locked:
            raise FUSEError(errno.EPERM)

        if not stat.S_ISDIR(inode.mode):
            raise FUSEError(errno.ENOTDIR)

        await self._remove(id_p, name, inode.id)

        await self.forget([(inode.id, 1)])

    async def _remove(self, id_p, name, id_, force=False):
        '''Remove entry `name` with parent inode `id_p`

        `id_` must be the inode of `name`. If `force` is True, then
        the `locked` attribute is ignored.
        '''

        log.debug('started with %d, %r', id_p, name)

        now_ns = time_ns()

        # Check that there are no child entries
        if self.db.has_val("SELECT 1 FROM contents WHERE parent_inode=?", (id_,)):
            log.debug("Attempted to remove entry with children: %s", get_path(id_p, self.db, name))
            raise FUSEError(errno.ENOTEMPTY)

        if self.inodes[id_p].locked and not force:
            raise FUSEError(errno.EPERM)

        name_id = self._del_name(name)
        self.db.execute("DELETE FROM contents WHERE name_id=? AND parent_inode=?", (name_id, id_p))

        inode = self.inodes[id_]
        inode.refcount -= 1
        inode.ctime_ns = now_ns

        inode_p = self.inodes[id_p]
        inode_p.mtime_ns = now_ns
        inode_p.ctime_ns = now_ns

        if inode.refcount == 0 and id_ not in self.open_inodes:
            log.debug('removing from cache')
            await self.cache.remove(id_, 0, int(math.ceil(inode.size / self.max_obj_size)))
            # Since the inode is not open, it's not possible that new blocks
            # get created at this point and we can safely delete the inode
            self.db.execute(
                'UPDATE names SET refcount = refcount - 1 WHERE '
                'id IN (SELECT name_id FROM ext_attributes WHERE inode=?)',
                (id_,),
            )
            self.db.execute(
                'DELETE FROM names WHERE refcount=0 AND '
                'id IN (SELECT name_id FROM ext_attributes WHERE inode=?)',
                (id_,),
            )
            self.db.execute('DELETE FROM ext_attributes WHERE inode=?', (id_,))
            self.db.execute('DELETE FROM symlink_targets WHERE inode=?', (id_,))
            del self.inodes[id_]

        log.debug('finished')

    async def symlink(self, id_p, name, target, ctx):
        log.debug('started with %d, %r, %r', id_p, name, target)

        if self.failsafe:
            raise FUSEError(errno.EPERM)

        mode = (
            stat.S_IFLNK
            | stat.S_IRUSR
            | stat.S_IWUSR
            | stat.S_IXUSR
            | stat.S_IRGRP
            | stat.S_IWGRP
            | stat.S_IXGRP
            | stat.S_IROTH
            | stat.S_IWOTH
            | stat.S_IXOTH
        )

        # Unix semantics require the size of a symlink to be the length
        # of its target. Therefore, we create symlink directory entries
        # with this size. If the kernel ever learns to open and read
        # symlinks directly, it will read the corresponding number of \0
        # bytes.
        inode = self._create(id_p, name, mode, ctx, size=len(target))
        self.db.execute(
            'INSERT INTO symlink_targets (inode, target) VALUES(?,?)', (inode.id, target)
        )
        self.open_inodes[inode.id] += 1
        return inode.entry_attributes()

    async def rename(self, id_p_old, name_old, id_p_new, name_new, flags, ctx):
        if flags:
            raise FUSEError(errno.ENOTSUP)

        log.debug('started with %d, %r, %d, %r', id_p_old, name_old, id_p_new, name_new)
        if name_new == CTRL_NAME or name_old == CTRL_NAME:
            log.warning(
                'Attempted to rename s3ql control file (%s -> %s)',
                get_path(id_p_old, self.db, name_old),
                get_path(id_p_new, self.db, name_new),
            )
            raise FUSEError(errno.EACCES)

        if self.failsafe or self.inodes[id_p_old].locked or self.inodes[id_p_new].locked:
            raise FUSEError(errno.EPERM)

        inode_old = self._lookup(id_p_old, name_old, ctx)

        try:
            inode_new = self._lookup(id_p_new, name_new, ctx)
        except FUSEError as exc:
            if exc.errno != errno.ENOENT:
                raise
            else:
                target_exists = False
        else:
            target_exists = True

        if target_exists:
            self._replace(id_p_old, name_old, id_p_new, name_new, inode_old.id, inode_new.id)
            await self.forget([(inode_old.id, 1), (inode_new.id, 1)])
        else:
            self._rename(id_p_old, name_old, id_p_new, name_new)
            await self.forget([(inode_old.id, 1)])

    def _add_name(self, name):
        '''Get id for *name* and increase refcount

        Name is inserted in table if it does not yet exist.
        '''

        try:
            name_id = self.db.get_val('SELECT id FROM names WHERE name=?', (name,))
        except NoSuchRowError:
            name_id = self.db.rowid('INSERT INTO names (name, refcount) VALUES(?,?)', (name, 1))
        else:
            self.db.execute('UPDATE names SET refcount=refcount+1 WHERE id=?', (name_id,))
        return name_id

    def _del_name(self, name):
        '''Decrease refcount for *name*

        Name is removed from table if refcount drops to zero. Returns the
        (possibly former) id of the name.
        '''

        (name_id, refcount) = self.db.get_row(
            'SELECT id, refcount FROM names WHERE name=?', (name,)
        )

        if refcount > 1:
            self.db.execute('UPDATE names SET refcount=refcount-1 WHERE id=?', (name_id,))
        else:
            self.db.execute('DELETE FROM names WHERE id=?', (name_id,))

        return name_id

    def _rename(self, id_p_old, name_old, id_p_new, name_new):
        now_ns = time_ns()

        name_id_new = self._add_name(name_new)
        name_id_old = self._del_name(name_old)

        self.db.execute(
            "UPDATE contents SET name_id=?, parent_inode=? WHERE name_id=? AND parent_inode=?",
            (name_id_new, id_p_new, name_id_old, id_p_old),
        )

        inode_p_old = self.inodes[id_p_old]
        inode_p_old.mtime_ns = now_ns
        inode_p_old.ctime_ns = now_ns

        inode_p_new = self.inodes[id_p_new]
        inode_p_new.mtime_ns = now_ns
        inode_p_new.ctime_ns = now_ns

    def _replace(self, id_p_old, name_old, id_p_new, name_new, id_old, id_new):
        now_ns = time_ns()

        if self.db.has_val("SELECT 1 FROM contents WHERE parent_inode=?", (id_new,)):
            log.info(
                "Attempted to overwrite entry with children: %s",
                get_path(id_p_new, self.db, name_new),
            )
            raise FUSEError(errno.EINVAL)

        # Replace target
        name_id_new = self.db.get_val('SELECT id FROM names WHERE name=?', (name_new,))
        self.db.execute(
            "UPDATE contents SET inode=? WHERE name_id=? AND parent_inode=?",
            (id_old, name_id_new, id_p_new),
        )

        # Delete old name
        name_id_old = self._del_name(name_old)
        self.db.execute(
            'DELETE FROM contents WHERE name_id=? AND parent_inode=?', (name_id_old, id_p_old)
        )

        inode_new = self.inodes[id_new]
        inode_new.refcount -= 1
        inode_new.ctime_ns = now_ns

        inode_p_old = self.inodes[id_p_old]
        inode_p_old.ctime_ns = now_ns
        inode_p_old.mtime_ns = now_ns

        inode_p_new = self.inodes[id_p_new]
        inode_p_new.ctime_ns = now_ns
        inode_p_new.mtime_ns = now_ns

        if inode_new.refcount == 0 and id_new not in self.open_inodes:
            self.cache.remove(id_new, 0, int(math.ceil(inode_new.size / self.max_obj_size)))
            # Since the inode is not open, it's not possible that new blocks
            # get created at this point and we can safely delete the inode
            self.db.execute(
                'UPDATE names SET refcount = refcount - 1 WHERE '
                'id IN (SELECT name_id FROM ext_attributes WHERE inode=?)',
                (id_new,),
            )
            self.db.execute('DELETE FROM names WHERE refcount=0')
            self.db.execute('DELETE FROM ext_attributes WHERE inode=?', (id_new,))
            self.db.execute('DELETE FROM symlink_targets WHERE inode=?', (id_new,))
            del self.inodes[id_new]

    async def link(self, id_, new_id_p, new_name, ctx):
        log.debug('started with %d, %d, %r', id_, new_id_p, new_name)

        if new_name == CTRL_NAME or id_ == CTRL_INODE:
            log.warning(
                'Attempted to create s3ql control file at %s', get_path(new_id_p, self.db, new_name)
            )
            raise FUSEError(errno.EACCES)

        now_ns = time_ns()
        inode_p = self.inodes[new_id_p]

        if inode_p.refcount == 0:
            log.warning('Attempted to create entry %s with unlinked parent %d', new_name, new_id_p)
            raise FUSEError(errno.EINVAL)

        if self.failsafe or inode_p.locked:
            raise FUSEError(errno.EPERM)

        inode_p.ctime_ns = now_ns
        inode_p.mtime_ns = now_ns

        self.db.execute(
            "INSERT INTO contents (name_id, inode, parent_inode) VALUES(?,?,?)",
            (self._add_name(new_name), id_, new_id_p),
        )
        inode = self.inodes[id_]
        inode.refcount += 1
        inode.ctime_ns = now_ns

        self.open_inodes[inode.id] += 1
        return inode.entry_attributes()

    async def setattr(self, id_, attr, fields, fh, ctx):
        """Handles FUSE setattr() requests"""
        inode = self.inodes[id_]
        if fh is not None:
            assert fh == id_
        now_ns = time_ns()

        if self.failsafe or inode.locked:
            raise FUSEError(errno.EPERM)

        if fields.update_mode:
            inode.mode = attr.st_mode

        if fields.update_uid:
            inode.uid = attr.st_uid

        if fields.update_gid:
            inode.gid = attr.st_gid

        if fields.update_atime:
            if attr.st_atime_ns.bit_length() > 63:
                raise FUSEError(errno.EINVAL)
            inode.atime_ns = attr.st_atime_ns

        if fields.update_mtime:
            if attr.st_mtime_ns.bit_length() > 63:
                raise FUSEError(errno.EINVAL)
            inode.mtime_ns = attr.st_mtime_ns

        inode.ctime_ns = now_ns

        # This needs to go last, because the call to cache.remove and cache.get
        # are asynchronous and may thus evict the *inode* object from the cache.
        if fields.update_size:
            len_ = attr.st_size

            # Determine blocks to delete
            last_block = len_ // self.max_obj_size
            cutoff = len_ % self.max_obj_size
            total_blocks = int(math.ceil(inode.size / self.max_obj_size))

            # Adjust file size
            inode.size = len_

            # Delete blocks and truncate last one if required
            if cutoff == 0:
                await self.cache.remove(id_, last_block, total_blocks)
            else:
                await self.cache.remove(id_, last_block + 1, total_blocks)
                try:
                    async with self.cache.get(id_, last_block) as fh:
                        fh.truncate(cutoff)
                except NoSuchObject as exc:
                    log.warning(
                        'Backend lost block %d of inode %d (id %s)!', last_block, id_, exc.key
                    )
                    raise

                except CorruptedObjectError as exc:
                    log.warning(
                        'Backend returned malformed data for block %d of inode %d (%s)',
                        last_block,
                        id_,
                        exc,
                    )
                    self.failsafe = True
                    self.broken_blocks[id_].add(last_block)
                    raise FUSEError(errno.EIO)

        return inode.entry_attributes()

    async def mknod(self, id_p, name, mode, rdev, ctx):
        log.debug('started with %d, %r', id_p, name)
        if self.failsafe:
            raise FUSEError(errno.EPERM)
        inode = self._create(id_p, name, mode, ctx, rdev=rdev)
        self.open_inodes[inode.id] += 1
        return inode.entry_attributes()

    async def mkdir(self, id_p, name, mode, ctx):
        log.debug('started with %d, %r', id_p, name)
        if self.failsafe:
            raise FUSEError(errno.EPERM)
        inode = self._create(id_p, name, mode, ctx)
        self.open_inodes[inode.id] += 1
        return inode.entry_attributes()

    def extstat(self):
        '''Return extended file system statistics'''

        log.debug('started')

        # Flush inode cache to get better estimate of total fs size
        self.inodes.flush()

        entries = self.db.get_val("SELECT COUNT(rowid) FROM contents")
        objects = self.db.get_val("SELECT COUNT(id) FROM objects")
        inodes = self.db.get_val("SELECT COUNT(id) FROM inodes")
        fs_size = self.db.get_val('SELECT SUM(size) FROM inodes') or 0
        dedup_size = self.db.get_val('SELECT SUM(length) FROM objects') or 0

        # Objects that are currently being uploaded/compressed have size == -1
        compr_size = self.db.get_val('SELECT SUM(phys_size) FROM objects WHERE phys_size > 0') or 0

        return struct.pack(
            'QQQQQQQQQQQQ',
            entries,
            objects,
            inodes,
            fs_size,
            dedup_size,
            compr_size,
            self.db.get_size(),
            *self.cache.get_usage()
        )

    async def statfs(self, ctx, _cache=[]):
        log.debug('started')

        stat_ = pyfuse3.StatvfsData()

        # Some applications call statfs() very often and repeatedly (VSCode). Since the SELECT
        # queries need to iterate over the full database, this is very expensive and slows other
        # filesystem users down. To avoid that, we cache results for a short time.
        if _cache and time.time() - _cache[3] < 30:
            (objects, inodes, size) = _cache[:3]
        else:
            objects = self.db.get_val("SELECT COUNT(id) FROM objects")
            inodes = self.db.get_val("SELECT COUNT(id) FROM inodes")
            size = self.db.get_val('SELECT SUM(length) FROM objects')
            _cache[:] = (objects, inodes, size, time.time())

        if size is None:
            size = 0

        # file system block size, i.e. the minimum amount of space that can
        # be allocated. This doesn't make much sense for S3QL, so we just
        # return the average size of stored blocks.
        stat_.f_frsize = max(4096, size // objects) if objects != 0 else 4096

        # This should actually be the "preferred block size for doing IO.  However, `df` incorrectly
        # interprets f_blocks, f_bfree and f_bavail in terms of f_bsize rather than f_frsize as it
        # should (according to statvfs(3)), so the only way to return correct values *and* have df
        # print something sensible is to set f_bsize and f_frsize to the same value. (cf.
        # http://bugs.debian.org/671490)
        stat_.f_bsize = stat_.f_frsize

        # size of fs in f_frsize units. Since backend is supposed to be unlimited,
        # always return a half-full filesystem, but at least 1 TB)
        fs_size = max(2 * size, 1024**4)

        stat_.f_blocks = fs_size // stat_.f_frsize
        stat_.f_bfree = (fs_size - size) // stat_.f_frsize
        stat_.f_bavail = stat_.f_bfree  # free for non-root

        total_inodes = max(2 * inodes, 1000000)
        stat_.f_files = total_inodes
        stat_.f_ffree = total_inodes - inodes
        stat_.f_favail = total_inodes - inodes  # free for non-root

        return stat_

    async def open(self, id_, flags, ctx):
        log.debug('started with %d', id_)
        if (flags & os.O_RDWR or flags & os.O_WRONLY) and (
            self.failsafe or self.inodes[id_].locked
        ):
            raise FUSEError(errno.EPERM)

        if flags & os.O_TRUNC:
            if not (flags & os.O_RDWR or flags & os.O_WRONLY):
                #  behaviour is not defined in POSIX, we opt for an error
                raise FUSEError(errno.EINVAL)
            attr = await self.getattr(id_, ctx)
            if stat.S_ISREG(attr.st_mode):
                attr.st_mtime_ns = time_ns()
                attr.st_size = 0
                await self.setattr(id_, attr, _TruncSetattrFields, None, ctx)
            elif stat.S_ISFIFO(attr.st_mode) or stat.S_ISBLK(attr.st_mode):
                #  silently ignore O_TRUNC when FIFO or terminal block device
                pass
            else:
                #  behaviour is not defined in POSIX, we opt for an error
                raise FUSEError(errno.EINVAL)

        return pyfuse3.FileInfo(fh=id_, keep_cache=True)

    async def access(self, id_, mode, ctx):
        '''Check if requesting process has `mode` rights on `inode`.

        This method always returns true, since it should only be called
        when permission checking is disabled (if permission checking is
        enabled, the `default_permissions` FUSE option should be set).
        '''
        # Yeah, could be a function and has unused arguments
        # pylint: disable=R0201,W0613

        log.debug('started with %d', id_)
        return True

    async def create(self, id_p, name, mode, flags, ctx):
        log.debug('started with id_p=%d, %s', id_p, name)
        if self.failsafe:
            raise FUSEError(errno.EPERM)

        try:
            id_ = self.db.get_val(
                "SELECT inode FROM contents_v WHERE name=? AND parent_inode=?", (name, id_p)
            )
        except NoSuchRowError:
            inode = self._create(id_p, name, mode, ctx)
        else:
            await self.open(id_, flags, ctx)
            inode = self.inodes[id_]

        self.open_inodes[inode.id] += 1
        return (pyfuse3.FileInfo(fh=inode.id), inode.entry_attributes())

    def _create(self, id_p, name, mode, ctx, rdev=0, size=0):
        if name == CTRL_NAME:
            log.warning(
                'Attempted to create s3ql control file at %s', get_path(id_p, self.db, name)
            )
            raise FUSEError(errno.EACCES)

        now_ns = time_ns()
        inode_p = self.inodes[id_p]

        if inode_p.locked:
            raise FUSEError(errno.EPERM)

        if inode_p.refcount == 0:
            log.warning('Attempted to create entry %s with unlinked parent %d', name, id_p)
            raise FUSEError(errno.EINVAL)
        inode_p.mtime_ns = now_ns
        inode_p.ctime_ns = now_ns

        if inode_p.mode & stat.S_ISGID:
            gid = inode_p.gid
            if stat.S_ISDIR(mode):
                mode |= stat.S_ISGID
        else:
            gid = ctx.gid
        inode = self.inodes.create_inode(
            mtime_ns=now_ns,
            ctime_ns=now_ns,
            atime_ns=now_ns,
            uid=ctx.uid,
            gid=gid,
            mode=mode,
            refcount=1,
            rdev=rdev,
            size=size,
        )

        self.db.execute(
            "INSERT INTO contents(name_id, inode, parent_inode) VALUES(?,?,?)",
            (self._add_name(name), inode.id, id_p),
        )

        return inode

    async def read(self, fh, offset, length):
        '''Read `size` bytes from `fh` at position `off`

        Unless EOF is reached, returns exactly `size` bytes.
        '''
        # log.debug('started with %d, %d, %d', fh, offset, length)
        buf = BytesIO()
        inode = self.inodes[fh]

        # Make sure that we don't read beyond the file size. This
        # should not happen unless direct_io is activated, but it's
        # cheap and nice for testing.
        size = inode.size
        length = min(size - offset, length)

        while length > 0:
            tmp = await self._readwrite(fh, offset, length=length)
            buf.write(tmp)
            length -= len(tmp)
            offset += len(tmp)

        # Inode may have expired from cache
        inode = self.inodes[fh]

        if inode.atime_ns < inode.ctime_ns or inode.atime_ns < inode.mtime_ns:
            inode.atime_ns = time_ns()

        return buf.getvalue()

    async def write(self, fh, offset, buf):
        '''Handle FUSE write requests.'''

        # log.debug('started with %d, %d, datalen=%d', fh, offset, len(buf))

        if self.failsafe or self.inodes[fh].locked:
            raise FUSEError(errno.EPERM)

        total = len(buf)
        minsize = offset + total
        while buf:
            written = await self._readwrite(fh, offset, buf=buf)
            offset += written
            buf = buf[written:]

        # Update file size if changed
        # Fuse does not ensure that we do not get concurrent write requests,
        # so we have to be careful not to undo a size extension made by
        # a concurrent write.
        now_ns = time_ns()
        inode = self.inodes[fh]
        inode.size = max(inode.size, minsize)
        inode.mtime_ns = now_ns
        inode.ctime_ns = now_ns

        return total

    async def _readwrite(self, id_, offset, *, buf=None, length=None):
        """Read or write as much as we can.

        If *buf* is None, read and return up to *length* bytes.

        If *length* is None, write from *buf* and return the number of
        bytes written.

        This is one method to reduce code duplication.
        """

        # Calculate required block
        blockno = offset // self.max_obj_size
        offset_rel = offset - blockno * self.max_obj_size

        if id_ in self.broken_blocks and blockno in self.broken_blocks[id_]:
            raise FUSEError(errno.EIO)

        if length is None:
            write = True
            length = len(buf)
        elif buf is None:
            write = False
            max_write = None
        else:
            raise TypeError("Don't know what to do!")

        # Don't try to write/read into the next block
        if offset_rel + length > self.max_obj_size:
            length = self.max_obj_size - offset_rel

        if write:
            # Point to the last byte that we will write
            max_write = offset_rel + length - 1
        else:
            max_write = None

        try:
            async with self.cache.get(id_, blockno, max_write=max_write) as fh:
                fh.seek(offset_rel)
                if write:
                    fh.write(buf[:length])
                else:
                    buf = fh.read(length)

        except NoSuchObject as exc:
            log.error('Backend lost block %d of inode %d (id %s)!', blockno, id_, exc.key)
            self.failsafe = True
            self.broken_blocks[id_].add(blockno)
            raise FUSEError(errno.EIO)

        except CorruptedObjectError as exc:
            log.error(
                'Backend returned malformed data for block %d of inode %d (%s)', blockno, id_, exc
            )
            self.failsafe = True
            self.broken_blocks[id_].add(blockno)
            raise FUSEError(errno.EIO)

        if write:
            return length
        elif len(buf) == length:
            return buf
        else:
            # If we can't read enough, add null bytes
            return buf + b"\0" * (length - len(buf))

    async def fsync(self, fh, datasync):
        log.debug('started with %d, %s', fh, datasync)
        if not datasync:
            self.inodes.flush_id(fh)

        for blockno in range(0, self.inodes[fh].size // self.max_obj_size + 1):
            self.cache.flush_local(fh, blockno)

    async def forget(self, forget_list):
        log.debug('started with %s', forget_list)

        for id_, nlookup in forget_list:
            self.open_inodes[id_] -= nlookup

            if self.open_inodes[id_] == 0:
                del self.open_inodes[id_]
                if id_ in self.broken_blocks:
                    del self.broken_blocks[id_]

                inode = self.inodes[id_]
                if inode.refcount == 0:
                    log.debug('removing %d from cache', id_)
                    await self.cache.remove(id_, 0, inode.size // self.max_obj_size + 1)
                    # Since the inode is not open, it's not possible that new blocks
                    # get created at this point and we can safely delete the inode
                    self.db.execute(
                        'UPDATE names SET refcount = refcount - 1 WHERE '
                        'id IN (SELECT name_id FROM ext_attributes WHERE inode=?)',
                        (id_,),
                    )
                    self.db.execute(
                        'DELETE FROM names WHERE refcount=0 AND '
                        'id IN (SELECT name_id FROM ext_attributes WHERE inode=?)',
                        (id_,),
                    )
                    self.db.execute('DELETE FROM ext_attributes WHERE inode=?', (id_,))
                    self.db.execute('DELETE FROM symlink_targets WHERE inode=?', (id_,))
                    del self.inodes[id_]

    async def fsyncdir(self, fh, datasync):
        log.debug('started with %d, %s', fh, datasync)
        if not datasync:
            self.inodes.flush_id(fh)

    async def releasedir(self, fh):
        log.debug('started with %d', fh)

    async def release(self, fh):
        log.debug('started with %d', fh)

    async def flush(self, fh):
        log.debug('started with %d', fh)


def update_logging(level, modules):
    root_logger = logging.getLogger()
    root_logger.setLevel(logging.INFO)
    if level == logging.DEBUG:
        logging.disable(logging.NOTSET)
        if 'all' in modules:
            root_logger.setLevel(logging.DEBUG)
        else:
            for module in modules:
                logging.getLogger(module).setLevel(logging.DEBUG)

    else:
        logging.disable(logging.DEBUG)



#!/usr/bin/env python3
'''
t5_failsafe.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

if __name__ == '__main__':
    import sys

    import pytest

    sys.exit(pytest.main([__file__] + sys.argv[1:]))

import errno
import os.path
import signal
import time
from argparse import Namespace

import pytest
import t4_fuse
from common import NoTestSection, get_remote_test_info

import s3ql.ctrl
from s3ql import BUFSIZE
from s3ql.backends import gs


@pytest.mark.usefixtures('pass_reg_output')
class TestFailsafe(t4_fuse.TestFuse):
    '''
    Test behavior with corrupted backend. In contrast to the tests
    in t3_fs_api, here we also make sure that remote connections
    are properly reset.
    '''

    def setup_method(self, method):
        super().setup_method(method)
        try:
            (backend_login, backend_pw, self.storage_url) = get_remote_test_info('gs-test')
        except NoTestSection as exc:
            super().teardown_method(method)
            pytest.skip(exc.reason)

        self.backend_login = backend_login
        self.backend_passphrase = backend_pw

        self.backend = gs.Backend(
            Namespace(
                storage_url=self.storage_url,
                backend_login=backend_login,
                backend_password=backend_pw,
                backend_options={},
            )
        )

    def test(self):
        self.mkfs(max_obj_size=10 * 1024**2)
        self.mount()

        fname1 = os.path.join(self.mnt_dir, 'file1')
        fname2 = os.path.join(self.mnt_dir, 'file2')

        # We need lots of data to keep the connection alive
        # and reproduce issue 424
        with open(fname1, 'wb') as fh:
            with open('/dev/urandom', 'rb') as src:
                for _ in range(5):
                    fh.write(src.read(BUFSIZE))
        s3ql.ctrl.main(['flushcache', self.mnt_dir])

        with open(fname2, 'w') as fh:
            fh.write('Hello, second world')
        s3ql.ctrl.main(['flushcache', self.mnt_dir])

        # Unmount required to avoid reading from kernel cache
        self.umount()

        # Modify
        (val, meta) = self.backend.fetch('s3ql_data_1')
        self.backend.store('s3ql_data_1', val[:500] + b'oops' + val[500:], meta)

        # Try to read

        self.mount()
        with pytest.raises(IOError) as exc_info:
            with open(fname1, 'rb') as fh:
                fh.read()
        assert exc_info.value.errno == errno.EIO
        self.reg_output(
            r'^ERROR: Backend returned malformed data for block 0 of inode \d+ .+$', count=1
        )

        # This should still work
        with open(fname2, 'rb') as fh:
            fh.read()

        # But this should not
        with pytest.raises(PermissionError):
            open(fname2, 'wb')
        self.reg_output(
            r'^ERROR: Backend returned malformed data for block 0 of inode \d+ .+$', count=1
        )

        # Printed during umount
        self.reg_output(r'^WARNING: File system errors encountered, marking for fsck.$', count=1)


class TestSigInt(t4_fuse.TestFuse):
    '''
    Make sure that we gracefully exit on SIGINT
    '''

    def test(self):
        self.mkfs()

        # Mount file system
        self.mount()

        # wait until pyfuse3 main loop runs
        time.sleep(2)

        # send SIGINT
        self.mount_process.send_signal(signal.SIGINT)

        # wait for clean unmount
        self.mount_process.wait(5)

        # we exited successfully?
        assert self.mount_process.returncode == 0

        self.fsck()



'''
backends/gs.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import hashlib
import json
import logging
import os
import re
import ssl
import threading
import urllib.parse
from ast import literal_eval
from base64 import b64decode, b64encode
from itertools import count
from typing import Any, BinaryIO, Dict, Optional

from s3ql.http import (
    BodyFollowing,
    CaseInsensitiveDict,
    ConnectionClosed,
    HTTPConnection,
    HTTPResponse,
    UnsupportedResponse,
    is_temp_network_error,
)

from ..common import OAUTH_CLIENT_ID, OAUTH_CLIENT_SECRET, copyfh
from ..logging import QuietError
from .common import (
    AbstractBackend,
    AuthenticationError,
    AuthorizationError,
    CorruptedObjectError,
    DanglingStorageURLError,
    NoSuchObject,
    get_proxy,
    get_ssl_context,
    retry,
)

try:
    import google.auth as g_auth
except ModuleNotFoundError:
    g_auth = None


log = logging.getLogger(__name__)


class ServerResponseError(Exception):
    '''Raised if the server response cannot be parsed.

    For HTTP errors (i.e., non 2xx response codes), RequestError should
    always be used instead (since in that case the response body can
    not be expected to have any specific format).
    '''

    def __init__(self, resp: HTTPResponse, error: str):
        self.resp = resp
        self.error = error

    def __str__(self):
        return '<ServerResponseError: %s>' % self.error


class RequestError(Exception):
    '''
    An error returned by the server.
    '''

    def __init__(
        self, code: str, reason: str, message: Optional[str] = None, body: Optional[str] = None
    ):
        super().__init__()
        self.code = code
        self.reason = reason
        self.message = message
        self.body = body

    def __str__(self) -> str:
        if self.message:
            return '<RequestError, code=%d, reason=%r, message=%r>' % (
                self.code,
                self.reason,
                self.message,
            )
        elif self.body:
            return '<RequestError, code=%d, reason=%r, with body data>' % (self.code, self.reason)
        else:
            return '<RequestError, code=%d, reason=%r>' % (self.code, self.reason)


class AccessTokenExpired(Exception):
    '''
    Raised if the access token has expired.
    '''


def _parse_error_response(resp, body: bytes) -> RequestError:
    '''Return exception corresponding to server response.'''

    try:
        json_resp = _parse_json_response(resp, body)
    except ServerResponseError:
        # Error messages may come from intermediate proxies and thus may not be in JSON.
        log.debug('Server response not JSON - intermediate proxy failure?')
        return RequestError(code=resp.status, reason=resp.reason)

    try:
        message = json_resp['error']['message']
        body = None
    except KeyError:
        log.warning('Did not find error.message element in JSON error response.')
        message = None
        body = str(json_resp)

    return RequestError(code=resp.status, reason=resp.reason, message=message)


def _parse_json_response(resp, body: bytes):
    # Note that even though the final server backend may guarantee to always deliver a JSON document
    # body with a detailed error message, we may also get errors from intermediate proxies.
    content_type = resp.headers.get('Content-Type', None)
    if content_type:
        hit = re.match(
            r'application/json(?:; charset="(.+)")?$',
            resp.headers['Content-Type'],
            re.IGNORECASE,
        )
    if not content_type or not hit:
        raise ServerResponseError(resp, error='expected json, got %s' % content_type)
    charset = hit.group(1)

    try:
        body_text = body.decode(charset)
    except UnicodeDecodeError as exc:
        log.warning('Unable to decode JSON response as Unicode (%s)', str(exc))
        raise ServerResponseError(resp, error=str(exc))

    try:
        resp_json = json.loads(body_text)
    except json.JSONDecodeError as exc:
        log.warning('Unable to decode JSON response (%s)', str(exc))
        raise ServerResponseError(resp, error=str(exc))

    return resp_json


class Backend(AbstractBackend):
    """A backend to store data in Google Storage"""

    known_options = {'ssl-ca-path', 'tcp-timeout'}

    # We don't want to request an access token for each instance,
    # because there is a limit on the total number of valid tokens.
    # This class variable holds the mapping from refresh tokens to
    # access tokens.
    access_token = dict()
    _refresh_lock = threading.Lock()
    adc = None

    def __init__(self, options):
        super().__init__()

        self.ssl_context = get_ssl_context(
            options.backend_options.get('ssl-ca-path', None)
        )  # type: Optional[ssl.Context]
        self.options = options.backend_options  # type: Dict[str, str]
        self.proxy = get_proxy(ssl=True)  # type: str
        self.login = options.backend_login  # type: str
        self.refresh_token = options.backend_password  # type: str

        if self.login == 'adc':
            if g_auth is None:
                raise QuietError('ADC authentication requires the google.auth module')
            elif self.adc is None:
                import google.auth.transport.urllib3
                import urllib3

                # Deliberately ignore proxy and SSL context when attempting
                # to connect to Compute Engine Metadata server.
                requester = google.auth.transport.urllib3.Request(urllib3.PoolManager())
                try:
                    credentials, _ = g_auth.default(
                        request=requester,
                        scopes=['https://www.googleapis.com/auth/devstorage.full_control'],
                    )
                except g_auth.exceptions.DefaultCredentialsError as exc:
                    raise QuietError('ADC found no valid credential sources: ' + str(exc))
                type(self).adc = (credentials, requester)
        elif self.login != 'oauth2':
            raise QuietError("Google Storage backend requires OAuth2 or ADC authentication")

        # Special case for unit testing against local mock server
        hit = re.match(
            r'^gs://!unittest!'
            r'([^/:]+)'  # Hostname
            r':([0-9]+)'  # Port
            r'/([^/]+)'  # Bucketname
            r'(?:/(.*))?$',  # Prefix
            options.storage_url,
        )
        if hit:
            self.hostname = hit.group(1)
            self.port = int(hit.group(2))
            self.bucket_name = hit.group(3)
            self.prefix = hit.group(4) or ''
        else:
            hit = re.match(r'^gs://([^/]+)(?:/(.*))?$', options.storage_url)
            if not hit:
                raise QuietError('Invalid storage URL', exitcode=2)

            self.bucket_name = hit.group(1)
            self.hostname = 'www.googleapis.com'
            self.prefix = hit.group(2) or ''
            self.port = 443

        self.conn = self._get_conn()
        self._check_bucket()

    @retry
    def _check_bucket(self):
        '''Check if bucket exists and/or credentials are correct.'''

        path = '/storage/v1/b/' + urllib.parse.quote(self.bucket_name, safe='')
        try:
            resp = self._do_request('GET', path)
        except RequestError as exc:
            if exc.code == 404:
                raise DanglingStorageURLError("Bucket '%s' does not exist" % self.bucket_name)
            exc = _map_request_error(exc, None)
            if exc:
                raise exc
            raise
        _parse_json_response(resp, self.conn.readall())

    def reset(self):
        if self.conn is not None and (self.conn.response_pending() or self.conn._out_remaining):
            log.debug('Resetting state of http connection %d', id(self.conn))
            self.conn.disconnect()

    def _get_conn(self):
        '''Return connection to server'''

        conn = HTTPConnection(
            self.hostname, self.port, proxy=self.proxy, ssl_context=self.ssl_context
        )
        conn.timeout = int(self.options.get('tcp-timeout', 20))
        return conn

    def is_temp_failure(self, exc):  # IGNORE:W0613
        if is_temp_network_error(exc) or isinstance(exc, ssl.SSLError):
            # We probably can't use the connection anymore, so disconnect (can't use reset, since
            # this would immediately attempt to reconnect, circumventing retry logic)
            self.conn.disconnect()
            return True

        elif isinstance(exc, RequestError) and (500 <= exc.code <= 599 or exc.code == 408):
            return True

        elif isinstance(exc, AccessTokenExpired):
            # Delete may fail if multiple threads encounter the same error
            self.access_token.pop(self.refresh_token, None)
            return True

        # Not clear at all what is happening here, but in doubt we retry
        elif isinstance(exc, ServerResponseError):
            return True

        if g_auth and isinstance(exc, g_auth.exceptions.TransportError):
            return True

        return False

    def _assert_empty_response(self, resp):
        '''Assert that current response body is empty'''

        buf = self.conn.read(2048)
        if not buf:
            return  # expected

        body = '\n'.join('%s: %s' % x for x in resp.headers.items())

        hit = re.search(r'; charset="(.+)"$', resp.headers.get('Content-Type', ''), re.IGNORECASE)
        if hit:
            charset = hit.group(1)
            body += '\n' + buf.decode(charset, errors='backslashreplace')

        log.warning('Expected empty response body, but got data - this is odd.')
        raise ServerResponseError(resp, error='expected empty response')

    @retry
    def delete(self, key):
        log.debug('started with %s', key)
        path = '/storage/v1/b/%s/o/%s' % (
            urllib.parse.quote(self.bucket_name, safe=''),
            urllib.parse.quote(self.prefix + key, safe=''),
        )
        try:
            resp = self._do_request('DELETE', path)
            self._assert_empty_response(resp)
        except RequestError as exc:
            exc = _map_request_error(exc, key)
            if isinstance(exc, NoSuchObject):
                pass
            elif exc:
                raise exc
            else:
                raise

    def list(self, prefix=''):
        prefix = self.prefix + prefix
        strip = len(self.prefix)
        page_token = None
        while True:
            (els, page_token) = self._list_page(prefix, page_token)
            for el in els:
                yield el[strip:]
            if page_token is None:
                break

    @retry
    def _list_page(self, prefix, page_token=None, batch_size=1000):
        # Limit maximum number of results since we read everything
        # into memory (because Python JSON doesn't have a streaming API)
        query_string = {'prefix': prefix, 'maxResults': str(batch_size)}
        if page_token:
            query_string['pageToken'] = page_token

        path = '/storage/v1/b/%s/o' % (urllib.parse.quote(self.bucket_name, safe=''),)

        try:
            resp = self._do_request('GET', path, query_string=query_string)
        except RequestError as exc:
            exc = _map_request_error(exc, None)
            if exc:
                raise exc
            raise
        json_resp = _parse_json_response(resp, self.conn.readall())
        page_token = json_resp.get('nextPageToken', None)

        if 'items' not in json_resp:
            assert page_token is None
            return ((), None)

        return ([x['name'] for x in json_resp['items']], page_token)

    @retry
    def lookup(self, key):
        log.debug('started with %s', key)
        return _unwrap_user_meta(self._get_gs_meta(key))

    def _get_gs_meta(self, key):
        path = '/storage/v1/b/%s/o/%s' % (
            urllib.parse.quote(self.bucket_name, safe=''),
            urllib.parse.quote(self.prefix + key, safe=''),
        )
        try:
            resp = self._do_request('GET', path)
        except RequestError as exc:
            exc = _map_request_error(exc, key)
            if exc:
                raise exc
            raise
        return _parse_json_response(resp, self.conn.readall())

    @retry
    def get_size(self, key):
        json_resp = self._get_gs_meta(key)
        return json_resp['size']

    def readinto_fh(self, key: str, fh: BinaryIO):
        '''Transfer data stored under *key* into *fh*, return metadata.

        The data will be inserted at the current offset. If a temporary error (as defined by
        `is_temp_failure`) occurs, the operation is retried.
        '''

        return self._readinto_fh(key, fh, fh.tell())

    @retry
    def _readinto_fh(self, key: str, fh: BinaryIO, off: int):
        gs_meta = self._get_gs_meta(key)
        metadata = _unwrap_user_meta(gs_meta)

        path = '/storage/v1/b/%s/o/%s' % (
            urllib.parse.quote(self.bucket_name, safe=''),
            urllib.parse.quote(self.prefix + key, safe=''),
        )
        try:
            self._do_request('GET', path, query_string={'alt': 'media'})
        except RequestError as exc:
            exc = _map_request_error(exc, key)
            if exc:
                raise exc
            raise

        copyfh(self.conn, fh)

        return metadata

    def write_fh(
        self,
        key: str,
        fh: BinaryIO,
        metadata: Optional[Dict[str, Any]] = None,
        len_: Optional[int] = None,
    ):
        '''Upload *len_* bytes from *fh* under *key*.

        The data will be read at the current offset. If *len_* is None, reads until the
        end of the file.

        If a temporary error (as defined by `is_temp_failure`) occurs, the operation is
        retried. Returns the size of the resulting storage object.
        '''

        off = fh.tell()
        if len_ is None:
            fh.seek(0, os.SEEK_END)
            len_ = fh.tell() - off
        return self._write_fh(key, fh, off, len_, metadata or {})

    @retry
    def _write_fh(self, key: str, fh: BinaryIO, off: int, len_: int, metadata: Dict[str, Any]):
        metadata = json.dumps(
            {
                'metadata': _wrap_user_meta(metadata),
                'name': self.prefix + key,
            }
        )

        # Google Storage uses Content-Length to read the object data, so we
        # don't have to worry about the boundary occurring in the object data.
        boundary = 'foo_bar_baz'
        headers = CaseInsensitiveDict()
        headers['Content-Type'] = 'multipart/related; boundary=%s' % boundary

        body_prefix = '\n'.join(
            (
                '--' + boundary,
                'Content-Type: application/json; charset=UTF-8',
                '',
                metadata,
                '--' + boundary,
                'Content-Type: application/octet-stream',
                '',
                '',
            )
        ).encode()
        body_suffix = ('\n--%s--\n' % boundary).encode()

        body_size = len(body_prefix) + len(body_suffix) + len_

        path = '/upload/storage/v1/b/%s/o' % (urllib.parse.quote(self.bucket_name, safe=''),)
        query_string = {'uploadType': 'multipart'}
        try:
            resp = self._do_request(
                'POST',
                path,
                query_string=query_string,
                headers=headers,
                body=BodyFollowing(body_size),
            )
        except RequestError as exc:
            exc = _map_request_error(exc, key)
            if exc:
                raise exc
            raise

        assert resp.status == 100
        fh.seek(off)

        try:
            self.conn.write(body_prefix)
            copyfh(fh, self.conn, len_)
            self.conn.write(body_suffix)
        except ConnectionClosed:
            # Server closed connection while we were writing body data -
            # but we may still be able to read an error response
            try:
                resp = self.conn.read_response()
            except ConnectionClosed:  # No server response available
                pass
            else:
                log.warning(
                    'Server broke connection during upload, signaled %d %s',
                    resp.status,
                    resp.reason,
                )
            # Re-raise first ConnectionClosed exception
            raise

        resp = self.conn.read_response()
        # If we're really unlucky, then the token has expired while we were uploading data.
        if resp.status == 401:
            self.conn.discard()
            raise AccessTokenExpired()
        elif resp.status != 200:
            exc = _parse_error_response(resp, self.conn.co_readall())
            raise _map_request_error(exc, key) or exc
        _parse_json_response(resp, self.conn.readall())

        return body_size

    def close(self):
        self.conn.disconnect()

    def __str__(self):
        return '<gs.Backend, name=%s, prefix=%s>' % (self.bucket_name, self.prefix)

    # This method uses a different HTTP connection than its callers, but shares
    # the same retry logic. It is therefore possible that errors with this
    # connection cause the other connection to be reset - but this should not
    # be a problem, because there can't be a pending request if we don't have
    # a valid access token.
    def _get_access_token(self):
        log.info('Requesting new access token')

        if self.adc:
            try:
                self.adc[0].refresh(self.adc[1])
            except g_auth.exceptions.RefreshError as exc:
                raise AuthenticationError('Failed to refresh credentials: ' + str(exc))
            self.access_token[self.refresh_token] = self.adc[0].token
            return

        headers = CaseInsensitiveDict()
        headers['Content-Type'] = 'application/x-www-form-urlencoded; charset=utf-8'

        body = urllib.parse.urlencode(
            {
                'client_id': OAUTH_CLIENT_ID,
                'client_secret': OAUTH_CLIENT_SECRET,
                'refresh_token': self.refresh_token,
                'grant_type': 'refresh_token',
            }
        )

        conn = HTTPConnection(
            'accounts.google.com', 443, proxy=self.proxy, ssl_context=self.ssl_context
        )
        try:
            conn.send_request('POST', '/o/oauth2/token', headers=headers, body=body.encode('utf-8'))
            resp = conn.read_response()
            json_resp = _parse_json_response(resp, conn.readall())

            if resp.status > 299 or resp.status < 200:
                assert 'error' in json_resp
            if 'error' in json_resp:
                raise AuthenticationError(json_resp['error'])
            else:
                self.access_token[self.refresh_token] = json_resp['access_token']
        finally:
            conn.disconnect()

    def _dump_body(self, resp):
        '''Return truncated string representation of response body.'''

        is_truncated = False
        try:
            body = self.conn.read(2048)
            if self.conn.read(1):
                is_truncated = True
                self.conn.discard()
        except UnsupportedResponse:
            log.warning('Unsupported response, trying to retrieve data from raw socket!')
            body = self.conn.read_raw(2048)
            self.conn.close()

        hit = re.search(r'; charset="(.+)"$', resp.headers.get('Content-Type', ''), re.IGNORECASE)
        if hit:
            charset = hit.group(1)
        else:
            charset = 'utf-8'

        body = body.decode(charset, errors='backslashreplace')

        if is_truncated:
            body += '... [truncated]'

        return body

    def _do_request(self, method, path, query_string=None, headers=None, body=None):
        '''Send request, read and return response object'''

        log.debug('started with %s %s, qs=%s', method, path, query_string)

        if headers is None:
            headers = CaseInsensitiveDict()

        expect100 = isinstance(body, BodyFollowing)
        headers['host'] = self.hostname
        if query_string:
            s = urllib.parse.urlencode(query_string, doseq=True)
            path += '?%s' % s

        # If we have an access token, try to use it.
        token = self.access_token.get(self.refresh_token, None)
        if token is not None:
            headers['Authorization'] = 'Bearer ' + token
            self.conn.send_request(method, path, body=body, headers=headers, expect100=expect100)
            resp = self.conn.read_response()
            if (expect100 and resp.status == 100) or (not expect100 and 200 <= resp.status <= 299):
                return resp
            elif resp.status != 401:
                raise _parse_error_response(resp, self.conn.readall())
            self.conn.discard()

        # If we reach this point, then the access token must have
        # expired, so we try to get a new one. We use a lock to prevent
        # multiple threads from refreshing the token simultaneously.
        with self._refresh_lock:
            # Don't refresh if another thread has already done so while
            # we waited for the lock.
            if token is None or self.access_token.get(self.refresh_token, None) == token:
                self._get_access_token()

        # Try request again. If this still fails, propagate the error
        # (because we have just refreshed the access token).
        # FIXME: We can't rely on this if e.g. the system hibernated
        # after refreshing the token, but before reaching this line.
        headers['Authorization'] = 'Bearer ' + self.access_token[self.refresh_token]
        self.conn.send_request(method, path, body=body, headers=headers, expect100=expect100)
        resp = self.conn.read_response()
        if (expect100 and resp.status == 100) or (not expect100 and 200 <= resp.status <= 299):
            return resp
        else:
            raise _parse_error_response(resp, self.conn.readall())


def _map_request_error(exc: RequestError, key: str):
    '''Map RequestError to more general exception if possible'''

    if exc.code == 404 and key:
        return NoSuchObject(key)
    elif exc.message == 'Forbidden':
        return AuthorizationError(exc.message)
    elif exc.message in ('Login Required', 'Invalid Credentials'):
        return AuthenticationError(exc.message)

    return None


def _wrap_user_meta(user_meta):
    obj_meta = dict()
    for key, val in user_meta.items():
        if not isinstance(key, str):
            raise TypeError('metadata keys must be str, not %s' % type(key))
        if not isinstance(val, (str, bytes, int, float, complex, bool)) and val is not None:
            raise TypeError('value for key %s (%s) is not elementary' % (key, val))
        if isinstance(val, (bytes, bytearray)):
            val = b64encode(val)

        obj_meta[key] = repr(val)

    return obj_meta


def _unwrap_user_meta(json_resp):
    '''Extract user metadata from JSON object metadata'''

    meta_raw = json_resp.get('metadata', None)
    if meta_raw is None:
        return {}

    # Detect Legacy format.
    if (
        meta_raw.get('format', None) == 'raw2'
        and 'md5' in meta_raw
        and all(key in ('format', 'md5') or re.match(r'^\d\d\d$', key) for key in meta_raw.keys())
    ):
        parts = []
        for i in count():
            part = meta_raw.get('%03d' % i, None)
            if part is None:
                break
            parts.append(part)
        buf = ''.join(parts)
        meta = literal_eval('{ %s }' % buf)
        for k, v in meta.items():
            if isinstance(v, bytes):
                meta[k] = b64decode(v)

        return meta

    meta = {}
    for k, v in meta_raw.items():
        try:
            v2 = literal_eval(v)
        except ValueError as exc:
            raise CorruptedObjectError('Invalid metadata value: ' + str(exc))
        if isinstance(v2, bytes):
            meta[k] = b64decode(v2)
        else:
            meta[k] = v2

    return meta


def md5sum_b64(buf):
    '''Return base64 encoded MD5 sum'''

    return b64encode(hashlib.md5(buf).digest()).decode('ascii')



'''
s3c4.py - this file is part of S3QL.

Copyright © 2008 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

import hashlib
import hmac
import logging
import time
import urllib.parse

from . import s3c

log = logging.getLogger(__name__)

# Maximum number of keys that can be deleted at once
MAX_KEYS = 1000

# Pylint goes berserk with false positives
# pylint: disable=E1002,E1101


class Backend(s3c.Backend):
    """A backend to stored data in some S3 compatible storage service.

    This classes uses AWS Signature V4 for authorization.
    """

    known_options = s3c.Backend.known_options | {'sig-region'}

    def __init__(self, options):
        self.sig_region = options.backend_options.get('sig-region', 'us-east-1')
        self.signing_key = None
        super().__init__(options)

    def __str__(self):
        return 's3c4://%s/%s/%s' % (self.hostname, self.bucket_name, self.prefix)

    def _authorize_request(self, method, path, headers, subres, query_string):
        '''Add authorization information to *headers*'''

        # See http://docs.aws.amazon.com/AmazonS3/latest/API/sigv4-auth-using-authorization-header.html

        now = time.gmtime()
        # now = time.strptime('Fri, 24 May 2013 00:00:00 GMT',
        #                    '%a, %d %b %Y %H:%M:%S GMT')

        ymd = time.strftime('%Y%m%d', now)
        ymdhms = time.strftime('%Y%m%dT%H%M%SZ', now)

        # add non-standard port to host header, needed for correct signature
        if self.port != 443:
            headers['host'] = '%s:%s' % (self.hostname, self.port)

        headers['x-amz-date'] = ymdhms
        headers['x-amz-content-sha256'] = 'UNSIGNED-PAYLOAD'

        headers.pop('Authorization', None)

        auth_strs = [method]
        auth_strs.append(urllib.parse.quote(path))

        if query_string:
            s = urllib.parse.urlencode(
                query_string, doseq=True, quote_via=urllib.parse.quote
            ).split('&')
        else:
            s = []
        if subres:
            s.append(urllib.parse.quote(subres) + '=')
        if s:
            s = '&'.join(sorted(s))
        else:
            s = ''
        auth_strs.append(s)

        # Headers
        sig_hdrs = sorted(x.lower() for x in headers.keys())
        for hdr in sig_hdrs:
            auth_strs.append('%s:%s' % (hdr, headers[hdr].strip()))
        auth_strs.append('')
        auth_strs.append(';'.join(sig_hdrs))
        auth_strs.append(headers['x-amz-content-sha256'])
        can_req = '\n'.join(auth_strs)
        # log.debug('canonical request: %s', can_req)

        can_req_hash = hashlib.sha256(can_req.encode()).hexdigest()
        str_to_sign = (
            "AWS4-HMAC-SHA256\n"
            + ymdhms
            + '\n'
            + '%s/%s/s3/aws4_request\n' % (ymd, self.sig_region)
            + can_req_hash
        )
        # log.debug('string to sign: %s', str_to_sign)

        if self.signing_key is None or self.signing_key[1] != ymd:
            self.update_signing_key(ymd)
        signing_key = self.signing_key[0]

        sig = hmac_sha256(signing_key, str_to_sign.encode(), hex=True)

        cred = '%s/%04d%02d%02d/%s/s3/aws4_request' % (
            self.login,
            now.tm_year,
            now.tm_mon,
            now.tm_mday,
            self.sig_region,
        )

        headers['Authorization'] = (
            'AWS4-HMAC-SHA256 '
            'Credential=%s,'
            'SignedHeaders=%s,'
            'Signature=%s' % (cred, ';'.join(sig_hdrs), sig)
        )

    def update_signing_key(self, ymd):
        date_key = hmac_sha256(("AWS4" + self.password).encode(), ymd.encode())
        region_key = hmac_sha256(date_key, self.sig_region.encode())
        service_key = hmac_sha256(region_key, b's3')
        signing_key = hmac_sha256(service_key, b'aws4_request')

        self.signing_key = (signing_key, ymd)


def hmac_sha256(key, msg, hex=False):
    d = hmac.new(key, msg, hashlib.sha256)
    if hex:
        return d.hexdigest()
    else:
        return d.digest()



#!/usr/bin/env python3
'''
t3_verify.py - this file is part of S3QL.

Copyright © 2014 Nikolaus Rath <Nikolaus@rath.org>

This work can be distributed under the terms of the GNU GPLv3.
'''

if __name__ == '__main__':
    import sys

    import pytest

    sys.exit(pytest.main([__file__] + sys.argv[1:]))

import io
import logging
import shutil
import tempfile
from argparse import Namespace

import pytest
from pytest_checklogs import assert_logs

from s3ql import verify
from s3ql.backends import local
from s3ql.backends.comprenc import ComprencBackend, sha256
from s3ql.database import Connection, create_tables
from s3ql.mkfs import init_tables


@pytest.fixture()
def backend():
    backend_dir = tempfile.mkdtemp(prefix='s3ql-backend-')
    plain_backend = local.Backend(Namespace(storage_url='local://' + backend_dir))
    backend = ComprencBackend(b'schnorz', ('zlib', 6), plain_backend)
    try:
        yield backend
    finally:
        backend.close()
        shutil.rmtree(backend_dir)


@pytest.fixture()
def db():
    dbfile = tempfile.NamedTemporaryFile()
    db = Connection(dbfile.name)
    create_tables(db)
    init_tables(db)
    try:
        yield db
    finally:
        db.close()
        dbfile.close()


@pytest.mark.parametrize("full", (True, False))
def test_missing(backend, db, full):
    # Create two objects, one will be missing
    obj_ids = (22, 25)
    missing_idx = 1
    for id_ in obj_ids:
        data = b'just some data that no-one really cares about %d' % id_
        db.execute(
            'INSERT INTO objects (id, refcount, phys_size, length, hash) VALUES(?, ?, ?, ?, ?)',
            (id_, 1, 27 * id_, len(data), sha256(data)),
        )

        if id_ != obj_ids[missing_idx]:
            key = 's3ql_data_%d' % obj_ids[0]
            backend[key] = data

    # When using a single thread, we can fake the backend factory
    backend_factory = lambda: backend

    missing_fh = io.StringIO()
    corrupted_fh = io.StringIO()
    with assert_logs('^Backend seems to have lost', count=1, level=logging.WARNING):
        verify.retrieve_objects(
            db, backend_factory, corrupted_fh, missing_fh, thread_count=1, full=full
        )
    assert missing_fh.getvalue() == 's3ql_data_%d\n' % obj_ids[missing_idx]
    assert corrupted_fh.getvalue() == ''


@pytest.mark.parametrize("full", (True, False))
def test_corrupted_head(backend, db, full):
    obj_ids = (30, 31)
    corrupted_idx = 1
    for id_ in obj_ids:
        data = b'just some data that no-one really cares about %d' % id_
        db.execute(
            'INSERT INTO objects (id, refcount, phys_size, length, hash) VALUES(?, ?, ?, ?, ?)',
            (id_, 1, 27 * id_, len(data), sha256(data)),
        )

        key = 's3ql_data_%d' % id_
        backend[key] = data

    # Introduce checksum error in metadata
    key = 's3ql_data_%d' % obj_ids[corrupted_idx]
    data = b'some data that will be broken on a metadata check'
    backend.store(
        key,
        data,
        {
            'meta-key1': 'some textual data that just increases',
            'meta-key2': 'the metadata size so that we can tamper with it',
        },
    )
    meta = backend.backend.lookup(key)
    raw = bytearray(meta['data'])
    assert len(raw) > 20
    raw[-10:-6] = b'forg'
    meta['data'] = raw
    backend.backend.store(key, data, meta)

    # When using a single thread, we can fake the backend factory
    backend_factory = lambda: backend

    missing_fh = io.StringIO()
    corrupted_fh = io.StringIO()
    with assert_logs('^Object %d is corrupted', count=1, level=logging.WARNING):
        verify.retrieve_objects(
            db, backend_factory, corrupted_fh, missing_fh, thread_count=1, full=full
        )
    assert missing_fh.getvalue() == ''
    assert corrupted_fh.getvalue() == 's3ql_data_%d\n' % obj_ids[1]


@pytest.mark.parametrize("full", (True, False))
def test_corrupted_body(backend, db, full):
    obj_ids = (35, 40)
    corrupted_idx = 1
    for id_ in obj_ids:
        data = b'just some data that no-one really cares about %d' % id_
        db.execute(
            'INSERT INTO objects (id, refcount, phys_size, length, hash) VALUES(?, ?, ?, ?, ?)',
            (id_, 1, 27 * id_, len(data), sha256(data)),
        )
        backend['s3ql_data_%d' % id_] = data

    # Introduce checksum error
    key = 's3ql_data_%d' % obj_ids[corrupted_idx]
    (raw, meta) = backend.backend.fetch(key)
    raw = bytearray(raw)
    assert len(raw) > 20
    raw[-10:-6] = b'forg'
    backend.backend.store(key, raw, meta)

    # When using a single thread, we can fake the backend factory
    backend_factory = lambda: backend

    missing_fh = io.StringIO()
    corrupted_fh = io.StringIO()

    if full:
        with assert_logs('^Object %d is corrupted', count=1, level=logging.WARNING):
            verify.retrieve_objects(
                db, backend_factory, corrupted_fh, missing_fh, thread_count=1, full=full
            )
            assert missing_fh.getvalue() == ''
            assert corrupted_fh.getvalue() == 's3ql_data_%d\n' % obj_ids[corrupted_idx]
    else:
        # Should not show up when looking just at HEAD
        verify.retrieve_objects(
            db, backend_factory, corrupted_fh, missing_fh, thread_count=1, full=full
        )
        assert missing_fh.getvalue() == ''
        assert corrupted_fh.getvalue() == ''


@pytest.mark.parametrize("full", (True, False))
def test_truncated_body(backend, db, full):
    id_ = 35
    data = b'just some data that no-one really cares about'
    db.execute(
        'INSERT INTO objects (id, refcount, phys_size, length) VALUES(?, ?, ?, ?)',
        (id_, 1, 27 * id_, len(data) + 1),
    )
    key = 's3ql_data_%d' % id_
    backend[key] = data

    # When using a single thread, we can fake the backend factory
    backend_factory = lambda: backend

    missing_fh = io.StringIO()
    corrupted_fh = io.StringIO()

    if full:
        with assert_logs('^Object %d is corrupted', count=1, level=logging.WARNING):
            verify.retrieve_objects(
                db, backend_factory, corrupted_fh, missing_fh, thread_count=1, full=full
            )
            assert missing_fh.getvalue() == ''
            assert corrupted_fh.getvalue() == 's3ql_data_%d\n' % id_
    else:
        # Should not show up when looking just at HEAD
        verify.retrieve_objects(
            db, backend_factory, corrupted_fh, missing_fh, thread_count=1, full=full
        )
        assert missing_fh.getvalue() == ''
        assert corrupted_fh.getvalue() == ''


def test_corrupted_hash(backend, db):
    obj_ids = (35, 40)
    corrupted_idx = 0
    for id_ in obj_ids:
        data = b'just some data that no-one really cares about %d' % id_
        db.execute(
            'INSERT INTO objects (id, refcount, phys_size, length, hash) VALUES(?, ?, ?, ?, ?)',
            (id_, 1, 27 * id_, len(data), sha256(data)),
        )
        backend['s3ql_data_%d' % id_] = data

    # Introduce checksum error
    db.execute('UPDATE objects SET hash=? WHERE id=?', (sha256(b'foobar'), obj_ids[corrupted_idx]))

    # When using a single thread, we can fake the backend factory
    backend_factory = lambda: backend

    missing_fh = io.StringIO()
    corrupted_fh = io.StringIO()

    with assert_logs('^Object %d is corrupted', count=1, level=logging.WARNING):
        verify.retrieve_objects(
            db, backend_factory, corrupted_fh, missing_fh, thread_count=1, full=True
        )
        assert missing_fh.getvalue() == ''
        assert corrupted_fh.getvalue() == 's3ql_data_%d\n' % obj_ids[corrupted_idx]
